; Simple REQUIRE system, superficially compatible with SLIB's.
; 2000-11-19 / lth
;
; Will load .sch and .scm files, and .fasl file if it exists and is newer
; than the source.

(define require)                        ; (require <symbol>)
(define program-vicinity)               ; A parameter
(define in-vicinity)                    ; String-append, really

(let ((load load))

  (define modules
    (let ((lib "/home/lth/lib/")
          (exp "/home/lth/net/larceny/Experimental/"))
      `((apropos                      ,lib "apropos")
        (array                        ,lib "array")
        (array-util                   ,lib "array-util")
        (box                          ,lib "box")
        (brackets                     ,lib "brackets")
        (comment                      ,lib "comment")
        (control                      ,lib "control")
        (coroutine                    ,lib "coroutine")
        (defmacro                     ,lib "defmacro")
        (engine                       ,lib "engine")
        (exec-comment                 ,lib "exec-comment")
        (format                       ,lib "format")
        (fortune                      ,lib "fortune")
        (fqueue                       ,lib "fqueue")
        (generator                    ,lib "generator")
        (list                         ,lib "list")
        (list-set                     ,lib "list-set")
        (match                        ,lib ("match" "match-syntax"))
        (number                       ,lib "number")
        (queue                        ,lib "queue")
        (random                       ,lib "random")
        (regexp                       ,lib "regexp")
        (string                       ,lib "string")
        (unify                        ,lib "unify")
        (unix                         ,lib "unix")
        (experimental/circular        ,lib ("read-circular" "print-circular"))
        (experimental/iosys           ,exp "iosys")
        (experimental/transcript      ,exp "transcript")
        (experimental/poll            ,exp "poll")
        (experimental/unix            ,exp "unix")
        (experimental/unix-descriptor ,exp "unix-descriptor")
        (experimental/socket          ,exp "socket")
        (experimental/tasking         ,exp "tasking")
        (experimental/tasking-unix    ,exp "tasking-unix")
        (experimental/nonblocking-console ,exp "nonblocking-console")
        )))

  (define provided '())

  (define (load-file fn)

    (define (load-file fn)
      (load fn)
      (display "; Loaded ")
      (display fn)
      (newline))

    (let ((fasl (string-append fn ".fasl"))
          (sch (string-append fn ".sch"))
          (scm (string-append fn ".scm")))
      (let ((src (cond ((and (file-exists? sch) (file-exists? scm))
                        (error "Require is finding both " sch 
                               " and " scm "."))
                       ((file-exists? scm) scm)
                       (else sch))))
        (cond ((and (file-exists? fasl) (file-exists? src))
               (if (file-newer? fasl src)
                   (load-file fasl)
                   (load-file src)))
              ((file-exists? src)
               (load-file src))
              ((file-exists? fasl)
               (load-file fasl))
              (else
               (error "Can't find either \"" src "\" or \"" fasl "\"."))))))

  (define (do-require key)
    (let ((probe (assq key modules)))
      (if probe
          (if (not (provided? key))
              (let ((done #f))
                (dynamic-wind
                 (lambda () 
                   (provide key))
                 (lambda () 
                   (call-without-interrupts 
                     (lambda ()
                       (parameterize ((program-vicinity (cadr probe)))
                         (if (list? (caddr probe))
                             (for-each (lambda (f)
                                         (load-file 
                                          (in-vicinity (program-vicinity) f)))
                                       (caddr probe))
                             (load-file
                              (in-vicinity (program-vicinity) (caddr probe)))))
                       (set! done #t)))
                   (unspecified))
                 (lambda ()
                   (if (not done)
                       (unprovide key))))))
          (error "REQUIRE: " key " is not a valid key."))))

  (define (provide key)
    (if (not (provided? key))
        (set! provided (cons key provided))))
  
  (define (unprovide key)
    (set! provided (remq key provided)))

  (define (provided? key)
    (memq key provided))

  (define prog-vicinity
    (let ((p "./"))
      (lambda args
        (cond ((null? args) p)
              ((null? (cdr args))
               (set! p (car args))
               p)
              (else
               (error "Wrong number of arguments to program-vicinity."))))))

  (set! require do-require)
  (set! program-vicinity prog-vicinity)
  (set! in-vicinity 
        (lambda (vicinity file)
          (string-append vicinity file)))
  'require)

; eof
