; Useful string processing procedures
; 2000-05-25 / lth

; (string-index "abra" #\a) => 0
; (string-index "abra" #\r) => 2
; (string-index "abra" #\c) => #f

(define (string-index s c)
  (let ((len (string-length s)))
    (let loop ((i 0))
      (cond ((= i len) #f)
            ((char=? c (string-ref s i)) i)
            (else (loop (+ i 1)))))))

(define (string-split s constituent?)
  (let ((limit (string-length s)))
    (let loop ((i 0) (words '()))
      (cond ((>= i limit) 
             (reverse words))
            ((constituent? (string-ref s i))
             (let ((start i))
               (let loop2 ((i (+ i 1)))
                 (if (and (< i limit) (constituent? (string-ref s i)))
                     (loop2 (+ i 1))
                     (loop (+ i 1) (cons (substring s start i) words))))))
            (else
             (loop (+ i 1) words))))))
  
; Not clever.

(define (substring-match s subs . rest)
  (let* ((start (if (null? rest) 0 (car rest)))
	 (ls    (string-length s))
	 (lu    (string-length subs))
	 (limit (- ls lu)))
    (let loop ((i start))
      (cond ((> i limit)
	     #f)
	    ((string=? subs (substring s i (+ i lu)))
	     i)
	    (else (loop (+ i 1)))))))

(define (substring-match-replace x match new)
  (let ((i (substring-match x match)))
    (if (not i)
        x
        (string-append (substring x 0 i)
                       new
                       (substring x 
                                  (+ i (string-length match)) 
                                  (string-length x))))))

; Thought experiment: string-quasiquote
;   #`"abc,{expr}cde" => (string-append "abc" (->string expr) "cde")
; eg
;   #`"abc,{(+ 1 2)}cde" => "abc3cde"
; This is like a limited but convenient FORMAT, or like string substitution
; in many shell languages.  Eg
;   (let ((x 10))
;     #`"this is x now: ,{x}")
;
; So: a STRING-QUASIQUOTE macro (easy enough) and a reader syntax.

; eof

