; Output processing             -*- mode: scheme; mode: font-lock -*-
; 2000-07-09 / lth
;
; WARNING: this code is now a real mess.
;
; BASIC-STATISTICS takes a list as produced by CRUNCH and flattens
; it, producing more useful statistics as described below.
;
; Bugs: 
; - BASIC-STATISTICS is ridiculously complex; a style based on
;   accumulator objects would probably have been much better.
; - GENERATE-LIST-SORTED-BY-BENCHMARK defies explanation.

(define no-warnings #f)                 ; Disable warnings for missing fields.

(define-syntax inc!
  (syntax-rules ()
    ((inc! ?place ?n)
     (set! ?place (+ ?place ?n)))
    ((inc! ?place)
     (set! ?place (+ ?place 1)))))

(define-syntax vinc!
  (syntax-rules ()
    ((vinc! ?place ?idx ?n)
     (vector-set! ?place ?idx (+ (vector-ref ?place ?idx) ?n)))
    ((vinc! ?place ?idx)
     (vector-set! ?place ?idx (+ (vector-ref ?place ?idx) 1)))))


; Compute basic statistics.
;
; (The following description does _not_ list all fields.)
;
; Return a list of association lists, one association list for each
; benchmark and system combination.  Each assoc list has the following
; keys; accumulators and averages are over all iterations:
; 
;   benchmark-name                      symbol
;   system-info                         (gctype option ...)
;   elapsed-first-average               number (milliseconds)
;   elapsed-all-average                 same
;   elapsed-butfirst-average            same
;   elapsed-mutator-all-average         same
;   elapsed-gc-first-average            same
;   elapsed-gc-all-average              same
;   elapsed-gc-butfirst-average         same
;   elapsed-gc-minor-all-average        same
;   elapsed-gc-major-all-average        same
;   elapsed-gc-minor-butfirst-average   same
;   elapsed-gc-major-butfirst-average   same
;   elapsed-gc-full-first-average       same
;   elapsed-gc-full-all-average         same
;   elapsed-gc-full-butfirst-average    same
;   elapsed-all-variance                same  (- max-elapsed min-elapsed)
;   cpu-all-average                     same
;   cpu-butfirst-average                same
;   cpu-first-average                   same
;   cpu-gc-all-average                  number (milliseconds)
;   cpu-gc-promote-all-average          same
;   cpu-gc-collect-all-average          same
;   cpu-gc-full-all-average             same
;   bytes-allocated-average             number (bytes)
;   bytes-reclaimed-average             same
;   bytes-marked-average                same
;   bytes-max-heap                      same
;   bytes-max-process                   same
;   remembered-set-entries-average      vector of number (entries)
;   remembered-objects-scanned-average  vector of number (bytes)
;   remembered-set-ssb-transactions-average
;   remembered-set-add-transactions-average
;   remembered-set-inner-scan-cost-average    number (milliseconds)
;   remembered-set-outer-scan-cost-average    number (milliseconds)
;   has-np-remset                       boolean
;   garbage-collections-average         number
;   swb-total-assignments-average       number  "swb" = simulated write barrier
;   swb-array-assignments-average       number
;   swb-filtered-remembered-average     number
;   swb-filtered-constant-average       number
;   swb-filtered-not-old-to-young-average  number
;   swb-transactions-average            number
;
; In addition there are averages over the sums of all iterations of a run,
; so these are in some sense per-run averages:
;
;   run-bytes-allocated-average
;   run-bytes-marked-average
;   run-garbage-collections-average
;   run-cpu-all-average
;   run-cpu-gc-all-average
;   run-cpu-gc-promote-all-average
;   run-cpu-gc-collect-all-average
;   run-cpu-gc-full-all-average
;   run-gc-full-count-all-average
;
; These are the average over the iterations of the run that has the
; lowest average value for total cpu time across the iterations:
;
;   cpu-best
;   cpu-gc-best
;   cpu-gc-promote-best
;   cpu-gc-collect-best
;   garbage-collections-best
;   major-collections-best
;   minor-collections-best
;   bytes-allocated-best
;   bytes-marked-best
;
; These are the sum over the iterations of the run that has the 
; lowest value for total cpu time summed across the iterations:
;
;   run-cpu-best
;   run-cpu-gc-best
;   run-cpu-gc-promote-best
;   run-cpu-gc-collect-best
;   run-garbage-collections-best
;   run-major-collections-best
;   run-minor-collections-best
;   run-bytes-allocated-best
;   run-bytes-marked-best
;
; If the runs failed, the list contains only benchmark name and system 
; information.
;
; If has-np-remset is #t then the NP extra remset statistics are the last
; entries in the remset vectors.
;
; (An extract/compute style might work better here?  See basic2.sch.)

(define INFINITY 1000000)

(define (basic-statistics list-of-systems)

  ; Keep these sorted alphabetically

  (define benchmark-name #f)
  (define bytes-allocated-sum 0)
  (define bytes-marked-sum 0)
  (define bytes-reclaimed-sum 0)
  (define count #i0)
  (define count-first #i0)
  (define cpu-all-sum 0)
  (define cpu-first-sum 0)
  (define cpu-gc-all-sum 0)
  (define cpu-gc-promote-all-sum 0)
  (define cpu-gc-collect-all-sum 0)
  (define cpu-gc-full-all-sum 0)
  (define elapsed-min INFINITY)
  (define elapsed-max 0)
  (define elapsed-first-sum 0)
  (define elapsed-all-sum 0)
  (define elapsed-mutator-all-sum 0)
  (define elapsed-gc-first-sum 0)
  (define elapsed-gc-all-sum 0)
  (define elapsed-gc-full-sum 0)
  (define elapsed-gc-full-first-sum 0)
  (define elapsed-gc-major-first-sum 0)
  (define elapsed-gc-major-all-sum 0)
  (define elapsed-gc-minor-first-sum 0)
  (define elapsed-gc-minor-all-sum 0)
  (define full-gccount-first-sum 0)
  (define full-gccount-sum 0)
  (define gccount-sum 0)
  (define has-np #f)
  (define major-gccount-sum 0)
  (define max-heap 0)
  (define max-remset 0)
  (define max-process 0)
  (define minor-gccount-sum 0)
  (define not-present '())
  (define number-of-iterations 0)
  (define remset-add-sums (make-vector 1000 0))
  (define remset-entries-sums (make-vector 1000 0))
  (define remset-inner-scancost-sum 0)
  (define remset-length 0)
  (define remset-outer-scancost-sum 0)
  (define remset-scanned-sums (make-vector 1000 0))
  (define remset-ssb-sums (make-vector 1000 0))
  (define some-failed #f)
  (define swb-total-sum 0)
  (define swb-array-sum 0)
  (define swb-filt-rem-sum 0)
  (define swb-filt-const-sum 0)
  (define swb-filt-xgen-sum 0)
  (define swb-filt-trans-sum 0)
  (define system-info #f)
  (define system-settings #f)

  (define run-bytes-allocated-sum 0)
  (define run-bytes-marked-sum 0)
  (define run-garbage-collections-sum 0)
  (define run-cpu-all-sum 0)
  (define run-cpu-gc-all-sum 0)
  (define run-cpu-gc-promote-all-sum 0)
  (define run-cpu-gc-collect-all-sum 0)
  (define run-cpu-gc-full-all-sum 0)
  (define run-full-collections-all-sum 0)
  (define run-count #i0)

  ; Sum across iterations for best sums and averages.

  (define cpu-best 0)
  (define cpu-user-best 0)                   ;*
  (define cpu-system-best 0)                 ;*
  (define cpu-gc-best 0)
  (define cpu-gc-promote-best 0)
  (define cpu-gc-collect-best 0)
  (define cpu-gc-full-best 0)
  (define garbage-collections-best 0)
  (define major-collections-best 0)
  (define minor-collections-best 0)
  (define full-collections-best 0)
  (define bytes-allocated-best 0)
  (define bytes-copied-best 0)
  (define bytes-marked-best 0)
  (define bytes-copied-promotion-best 0)      
  (define bytes-copied-collection-best 0)     
  (define bytes-moved-best 0)                 
  (define bytes-marked-full-best 0)
  (define remset-entries-added-best 0)        
  (define remset-entries-scanned-best 0)      
  (define remset-entries-removed-best 0)      
  (define remset-bytes-scanned-best 0)        

  (define run-cpu-best INFINITY)
  (define run-cpu-user-best 0)                
  (define run-cpu-system-best 0)              
  (define run-cpu-gc-best 0)
  (define run-cpu-gc-promote-best 0)
  (define run-cpu-gc-collect-best 0)
  (define run-cpu-gc-full-best 0)
  (define run-garbage-collections-best 0)
  (define run-major-collections-best 0)
  (define run-minor-collections-best 0)
  (define run-full-collections-best 0)
  (define run-bytes-allocated-best 0)
  (define run-bytes-marked-best 0)
  (define run-bytes-copied-promotion-best 0)
  (define run-bytes-copied-collection-best 0)
  (define run-bytes-copied-best 0)
  (define run-bytes-moved-best 0)             
  (define run-bytes-marked-full-best 0)
  (define run-remset-entries-added-best 0)    
  (define run-remset-entries-scanned-best 0)  
  (define run-remset-entries-removed-best 0)  
  (define run-remset-bytes-scanned-best 0)    
  (define run-remset-trace-best '())
  (define run-frames-flushed-best 0)
  (define run-frames-restored-best 0)

  (define (crank-remset v)
    (list->vector
     (map (lambda (x) (/ x count))
          (list-head (vector->list v) remset-length))))

  (define (init-run-accumulators)
    (set! run-bytes-allocated-sum 0)
    (set! run-bytes-marked-sum 0)
    (set! run-garbage-collections-sum 0)
    (set! run-cpu-all-sum 0)
    (set! run-cpu-gc-all-sum 0)
    (set! run-cpu-gc-promote-all-sum 0)
    (set! run-cpu-gc-collect-all-sum 0)
    (set! run-cpu-gc-full-all-sum 0)
    (set! run-full-collections-all-sum 0)
    (set! run-count #i0)

    (set! run-cpu-best INFINITY)
    (set! run-cpu-gc-best 0)
    (set! run-cpu-user-best 0)
    (set! run-cpu-system-best 0)
    (set! run-cpu-gc-promote-best 0)
    (set! run-cpu-gc-collect-best 0)
    (set! run-cpu-gc-full-best 0)
    (set! run-garbage-collections-best 0)
    (set! run-major-collections-best 0)
    (set! run-minor-collections-best 0)
    (set! run-full-collections-best 0)
    (set! run-bytes-allocated-best 0)
    (set! run-bytes-marked-best 0)
    (set! run-bytes-copied-promotion-best 0)
    (set! run-bytes-copied-collection-best 0)
    (set! run-bytes-moved-promotion-best 0)
    (set! run-bytes-moved-collection-best 0)
    (set! run-bytes-copied-best 0)
    (set! run-bytes-moved-best 0)
    (set! run-bytes-marked-full-best 0)
    (set! run-remset-entries-added-best 0)
    (set! run-remset-entries-scanned-best 0)
    (set! run-remset-entries-removed-best 0)
    (set! run-remset-bytes-scanned-best 0)
    (set! run-iteration-count-best 0)
    (set! run-remset-trace-best '())
    (set! run-frames-flushed-best 0)
    (set! run-frames-restored-best 0))
  
  (define (init-iteration-accumulators)
    (set! benchmark-name #f)
    (set! bytes-allocated-sum 0)
    (set! bytes-marked-sum 0)
    (set! bytes-reclaimed-sum 0)
    (set! count #i0)
    (set! count-first #i0)
    (set! cpu-all-sum 0)
    (set! cpu-first-sum 0)
    (set! cpu-gc-all-sum 0)
    (set! cpu-gc-promote-all-sum 0)
    (set! cpu-gc-collect-all-sum 0)
    (set! cpu-gc-full-all-sum 0)
    (set! elapsed-min INFINITY)
    (set! elapsed-max 0)
    (set! elapsed-first-sum 0)
    (set! elapsed-all-sum 0)
    (set! elapsed-mutator-all-sum 0)
    (set! elapsed-gc-first-sum 0)
    (set! elapsed-gc-all-sum 0)
    (set! elapsed-gc-major-first-sum 0)
    (set! elapsed-gc-major-all-sum 0)
    (set! elapsed-gc-minor-first-sum 0)
    (set! elapsed-gc-minor-all-sum 0)
    (set! elapsed-gc-full-sum 0)
    (set! elapsed-gc-full-first-sum 0)
    (set! full-gccount-first-sum 0)
    (set! full-gccount-sum 0)
    (set! gccount-sum 0)
    (set! has-np #f)
    (set! major-gccount-sum 0)
    (set! max-heap 0)
    (set! max-remset 0)
    (set! max-process 0)
    (set! minor-gccount-sum 0)
    (set! number-of-iterations 0)
    (set! remset-entries-sums (make-vector 1000 0))
    (set! remset-scanned-sums (make-vector 1000 0))
    (set! remset-ssb-sums (make-vector 1000 0))
    (set! remset-add-sums (make-vector 1000 0))
    (set! remset-inner-scancost-sum 0)
    (set! remset-outer-scancost-sum 0)
    (set! remset-length 0)
    (set! some-failed #f)
    (set! swb-total-sum 0)
    (set! swb-array-sum 0)
    (set! swb-filt-rem-sum 0)
    (set! swb-filt-const-sum 0)
    (set! swb-filt-xgen-sum 0)
    (set! swb-filt-trans-sum 0)
    (set! system-info #f)
    (set! system-settings #f))

  (define (compute-averages)
    (if (> count 0.0)
        `((benchmark-name . ,benchmark-name)
          (system-info . ,system-info)
          (system-settings . ,system-settings)
          (elapsed-first-average . ,(/ elapsed-first-sum count-first))
          (elapsed-all-average . ,(/ elapsed-all-sum count))
          (elapsed-butfirst-average
           . ,(/ (- elapsed-all-sum elapsed-first-sum) (- count count-first)))
          (elapsed-mutator-all-average . ,(/ elapsed-mutator-all-sum count))
          (cpu-first-average . ,(/ cpu-first-sum count-first))
          (cpu-all-average . ,(/ cpu-all-sum count))
          (cpu-butfirst-average
           . ,(/ (- cpu-all-sum cpu-first-sum) (- count count-first)))
          (elapsed-gc-first-average . ,(/ elapsed-gc-first-sum count-first))
          (elapsed-gc-all-average . ,(/ elapsed-gc-all-sum count))
          (elapsed-gc-butfirst-average 
           . ,(/ (- elapsed-gc-all-sum elapsed-gc-first-sum) 
                 (- count count-first)))
          (elapsed-gc-full-first-average 
           . ,(/ elapsed-gc-full-first-sum count-first))
          (elapsed-gc-full-all-average . ,(/ elapsed-gc-full-sum count))
          (elapsed-gc-full-butfirst-average 
           . ,(/ (- elapsed-gc-full-sum elapsed-gc-full-first-sum)
                 (- count count-first)))
          (elapsed-gc-minor-all-average . ,(/ elapsed-gc-minor-all-sum count))
          (elapsed-gc-major-all-average . ,(/ elapsed-gc-major-all-sum count))
          (elapsed-gc-minor-butfirst-average
           . ,(/ (- elapsed-gc-minor-all-sum elapsed-gc-minor-first-sum)
                 (- count count-first)))
          (elapsed-gc-major-butfirst-average 
           . ,(/ (- elapsed-gc-major-all-sum elapsed-gc-major-first-sum)
                 (- count count-first)))
          (bytes-allocated-average . ,(/ bytes-allocated-sum count))
          (bytes-reclaimed-average . ,(/ bytes-reclaimed-sum count))
          (bytes-marked-average . ,(/ bytes-marked-sum count))
          (cpu-gc-all-average . ,(/ cpu-gc-all-sum count))
          (cpu-gc-promote-all-average . ,(/ cpu-gc-promote-all-sum count))
          (cpu-gc-collect-all-average . ,(/ cpu-gc-collect-all-sum count))
          (cpu-gc-full-all-average . ,(/ cpu-gc-full-all-sum count))
          (bytes-max-heap . ,max-heap)
          (bytes-max-remset . ,max-remset)
          (elapsed-all-variance . ,(- elapsed-max elapsed-min))
          (some-failed . ,some-failed)
          (garbage-collections-average . ,(/ gccount-sum count))
          (has-np-remset . ,has-np)
          (bytes-max-process . ,max-process)
          (number-of-iterations . ,number-of-iterations)
          (remembered-set-entries-average 
           . ,(crank-remset remset-entries-sums))
          (remembered-objects-scanned-average 
           . ,(crank-remset remset-scanned-sums))
          (remembered-set-ssb-transactions-average 
           . ,(crank-remset remset-ssb-sums))
          (remembered-set-add-transactions-average 
           . ,(crank-remset remset-add-sums))
          (remembered-set-inner-scan-cost-average 
           . ,(/ (/ remset-inner-scancost-sum 1000000.0) count))
          (remembered-set-outer-scan-cost-average 
           . ,(/ (/ remset-outer-scancost-sum 1000000.0) count))
          (swb-total-assignments-average . ,(/ swb-total-sum count))
          (swb-array-assignments-average . ,(/ swb-array-sum count))
          (swb-filtered-remembered-average . ,(/ swb-filt-rem-sum count))
          (swb-filtered-constant-average . ,(/ swb-filt-const-sum count))
          (swb-filtered-not-old-to-young-average 
           . ,(/ swb-filt-xgen-sum count))
          (swb-transactions-average . ,(/ swb-filt-trans-sum count))
          (run-bytes-allocated-average 
           . ,(/ run-bytes-allocated-sum run-count))
          (run-bytes-marked-average
           . ,(/ run-bytes-marked-sum run-count))
          (run-garbage-collections-average
           . ,(/ run-garbage-collections-sum run-count))
          (run-cpu-all-average
           . ,(/ run-cpu-all-sum run-count))
          (run-cpu-gc-all-average
           . ,(/ run-cpu-gc-all-sum run-count))
          (run-cpu-gc-promote-all-average
           . ,(/ run-cpu-gc-promote-all-sum run-count))
          (run-cpu-gc-collect-all-average
           . ,(/ run-cpu-gc-collect-all-sum run-count))
          (run-cpu-gc-full-all-average
           . ,(/ run-cpu-gc-full-all-sum run-count))
          (run-gc-full-count-all-average
           . ,(/ run-full-collections-all-sum run-count))

          (run-garbage-collections-best . ,run-garbage-collections-best)
          (run-bytes-allocated-best . ,run-bytes-allocated-best)
          (run-bytes-marked-best . ,run-bytes-marked-best)
          (run-cpu-best . ,run-cpu-best)
          (run-cpu-user-best . ,run-cpu-user-best)
          (run-cpu-system-best . ,run-cpu-system-best)
          (run-cpu-gc-best . ,run-cpu-gc-best)
          (run-cpu-gc-promote-best . ,run-cpu-gc-promote-best)
          (run-cpu-gc-collect-best . ,run-cpu-gc-collect-best)
          (run-cpu-gc-major-best . ,run-cpu-gc-major-best)
          (run-cpu-gc-minor-best . ,run-cpu-gc-minor-best)
          (run-cpu-gc-full-best . ,run-cpu-gc-full-best) 
          (run-major-collections-best . ,run-major-collections-best)
          (run-minor-collections-best . ,run-minor-collections-best)
          (run-full-collections-best . ,run-full-collections-best)
          (run-bytes-copied-promotion-best . ,run-bytes-copied-promotion-best)
          (run-bytes-copied-collection-best 
           . ,run-bytes-copied-collection-best)
          (run-bytes-moved-promotion-best . ,run-bytes-moved-promotion-best)
          (run-bytes-moved-collection-best 
           . ,run-bytes-moved-collection-best)
          (run-bytes-copied-best . ,run-bytes-copied-best)
          (run-bytes-moved-best . ,run-bytes-moved-best)
          (run-bytes-marked-full-best . ,run-bytes-marked-full-best)
          (run-remset-entries-added-best . ,run-remset-entries-added-best)
          (run-remset-entries-scanned-best . ,run-remset-entries-scanned-best)
          (run-remset-entries-removed-best . ,run-remset-entries-removed-best)
          (run-remset-bytes-scanned-best . ,run-remset-bytes-scanned-best)
          (run-remset-trace-best . ,run-remset-trace-best)
          (run-frames-flushed-best . ,run-frames-flushed-best)
          (run-frames-restored-best . ,run-frames-restored-best)

          (garbage-collections-best . ,garbage-collections-best)
          (bytes-allocated-best . ,bytes-allocated-best)
          (bytes-marked-best . ,bytes-marked-best)
          (cpu-best . ,cpu-best)
          (cpu-user-best . ,cpu-user-best)
          (cpu-system-best . ,cpu-system-best)
          (cpu-gc-best . ,cpu-gc-best)
          (cpu-gc-promote-best . ,cpu-gc-promote-best)
          (cpu-gc-collect-best . ,cpu-gc-collect-best)
          (cpu-gc-full-best . ,cpu-gc-full-best)
          (cpu-gc-major-best . ,cpu-gc-major-best)
          (cpu-gc-minor-best . ,cpu-gc-minor-best)
          (major-collections-best . ,major-collections-best)
          (minor-collections-best . ,minor-collections-best)
          (full-collections-best . ,full-collections-best)
          (bytes-moved-promotion-best . ,bytes-moved-promotion-best)
          (bytes-moved-collection-best . ,bytes-moved-collection-best)
          (bytes-copied-promotion-best . ,bytes-copied-promotion-best)
          (bytes-copied-collection-best . ,bytes-copied-collection-best)
          (bytes-copied-best . ,bytes-copied-best)
          (bytes-moved-best . ,bytes-moved-best)
          (bytes-marked-full-best . ,bytes-marked-full-best)
          (remset-entries-added-best . ,remset-entries-added-best)
          (remset-entries-scanned-best . ,remset-entries-scanned-best)
          (remset-entries-removed-best . ,remset-entries-removed-best)
          (remset-bytes-scanned-best . ,remset-bytes-scanned-best)
          )
        `((benchmark-name . ,benchmark-name)
          (system-info . ,system-info)
          (system-settings . ,system-settings)
          (some-failed . #t))))

  (define (do-one-run run)

    (define iteration 0)

    ; Per-run accumulators

    (define run-bytes-allocated-acc 0)
    (define run-garbage-collections-acc 0)
    (define run-cpu-acc 0)
    (define run-cpu-user-acc 0)
    (define run-cpu-system-acc 0)
    (define run-cpu-gc-acc 0)
    (define run-cpu-gc-promote-acc 0)
    (define run-cpu-gc-collect-acc 0)
    (define run-cpu-gc-minor-acc 0)
    (define run-cpu-gc-major-acc 0)
    (define run-cpu-gc-full-acc 0)
    (define run-minor-collections-acc 0)
    (define run-major-collections-acc 0)
    (define run-full-collections-acc 0)
    (define run-bytes-copied-acc 0)
    (define run-bytes-marked-acc 0)
    (define run-bytes-copied-promotion-acc 0)
    (define run-bytes-copied-collection-acc 0)
    (define run-bytes-moved-promotion-acc 0)
    (define run-bytes-moved-collection-acc 0)
    (define run-bytes-moved-acc 0)
    (define run-bytes-marked-full-acc 0)
    (define run-remset-entries-added-acc 0)
    (define run-remset-entries-scanned-acc 0)
    (define run-remset-entries-removed-acc 0)
    (define run-remset-bytes-scanned-acc 0)
    (define run-frames-flushed-acc 0)
    (define run-frames-restored-acc 0)

    (define (do-one-iter iter scancost)

      (define (get x) 
        (let ((probe (assq x iter)))
          (if probe 
              (cdr probe)
              (begin
                (if (not (memq x not-present))
                    (begin
                      (if (not no-warnings)
                          (format #t 
                                  "Warning: statistics field ~a not present.~%"
                                  x))
                      (set! not-present (cons x not-present))))
                0))))

      (inc! count)
      (inc! iteration)
      (if (= 1 iteration)
          (begin 
            (inc! cpu-first-sum 
                  (+ (get 'cpu-time-user) (get 'cpu-time-system)))
            (inc! elapsed-first-sum (get 'elapsed-time))
            (inc! elapsed-gc-first-sum (get 'elapsed-time-gc))
            (inc! elapsed-gc-full-first-sum (get 'elapsed-time-full-gc))
            (inc! full-gccount-first-sum (+ (get 'count-full-gc)))
            (inc! elapsed-gc-minor-first-sum (get 'elapsed-time-minor-gc))
            (inc! elapsed-gc-major-first-sum (get 'elapsed-time-major-gc))
            (inc! count-first)))
      (set! elapsed-max (max elapsed-max (get 'elapsed-time)))
      (set! elapsed-min (min elapsed-min (get 'elapsed-time)))
      (inc! cpu-all-sum (+ (get 'cpu-time-user) (get 'cpu-time-system)))
      (inc! cpu-gc-all-sum (get 'cpu-time-gc))
      (inc! cpu-gc-promote-all-sum (get 'cpu-time-gc-promotion))
      (inc! cpu-gc-collect-all-sum (get 'cpu-time-gc-collection))
      (inc! cpu-gc-full-all-sum (get 'cpu-time-gc-full))
      (inc! elapsed-all-sum (get 'elapsed-time))
      (inc! elapsed-mutator-all-sum 
            (- (get 'elapsed-time) (get 'elapsed-time-gc)))
      (inc! elapsed-gc-all-sum (get 'elapsed-time-gc))
      (inc! elapsed-gc-minor-all-sum (get 'elapsed-time-minor-gc))
      (inc! elapsed-gc-major-all-sum (get 'elapsed-time-major-gc))
      (inc! elapsed-gc-full-sum (get 'elapsed-time-full-gc))
      (inc! bytes-allocated-sum (get 'bytes-allocated))
      (inc! bytes-reclaimed-sum (get 'bytes-reclaimed))
      (inc! bytes-marked-sum (get 'bytes-marked))
      (inc! gccount-sum (+ (get 'count-gc)
                           (get 'count-full-gc)))
      (inc! major-gccount-sum (get 'count-major-gc))
      (inc! minor-gccount-sum (get 'count-minor-gc))
      (inc! full-gccount-sum (get 'count-full-gc))
      (do ((i 0 (+ i 1))
           (e (get 'remset-entries-scanned) (cdr e))
           (s (get 'remset-bytes-scanned) (cdr s))
           (ssb (get 'remset-ssb-transactions) (cdr ssb))
           (add (get 'remset-add-transactions) (cdr add)))
          ((null? e)
           (set! remset-length i))
        (vinc! remset-entries-sums i (car e))
        (vinc! remset-scanned-sums i (car s))
        (vinc! remset-ssb-sums i (car ssb))
        (vinc! remset-add-sums i (car add)))
      (set! has-np (get 'has-np-remset))
      (inc! remset-inner-scancost-sum (car scancost))
      (inc! remset-outer-scancost-sum (cadr scancost))
;      (inc! swb-total-sum (get 'swb-total-assignments))
;      (inc! swb-array-sum (get 'swb-array-assignments))
;      (inc! swb-filt-rem-sum (get 'swb-filtered-remembered))
;      (inc! swb-filt-const-sum (get 'swb-filtered-constant))
;      (inc! swb-filt-xgen-sum (get 'swb-filtered-not-old-to-young))
;      (inc! swb-filt-trans-sum (get 'swb-transactions))
      (set! max-heap (max max-heap (get 'bytes-max-heap-used)))
      (set! max-remset (max max-remset (get 'bytes-max-remset-used)))

      (inc! run-bytes-allocated-acc (get 'bytes-allocated))
      (inc! run-garbage-collections-acc 
            (+ (get 'count-gc) (get 'count-full-gc)))
      (inc! run-cpu-acc (+ (get 'cpu-time-user) (get 'cpu-time-system)))
      (inc! run-cpu-user-acc (get 'cpu-time-user))
      (inc! run-cpu-system-acc (get 'cpu-time-system))
      (inc! run-cpu-gc-acc (get 'cpu-time-gc))
      (inc! run-cpu-gc-promote-acc (get 'cpu-time-gc-promotion))
      (inc! run-cpu-gc-minor-acc (get 'cpu-time-gc-minor))
      (inc! run-cpu-gc-major-acc (get 'cpu-time-gc-major))
      (inc! run-cpu-gc-collect-acc (get 'cpu-time-gc-collection))
      (inc! run-cpu-gc-full-acc (get 'cpu-time-gc-full))
      (inc! run-minor-collections-acc (get 'count-minor-gc))
      (inc! run-major-collections-acc (get 'count-major-gc))
      (inc! run-full-collections-acc (get 'count-full-gc))
      (inc! run-bytes-copied-acc (get 'bytes-copied))
      (inc! run-bytes-marked-acc (get 'bytes-marked))
      (inc! run-bytes-copied-promotion-acc (get 'bytes-copied-promotion))
      (inc! run-bytes-copied-collection-acc (get 'bytes-copied-collection))
      (inc! run-bytes-moved-promotion-acc (get 'bytes-moved-promotion))
      (inc! run-bytes-moved-collection-acc (get 'bytes-moved-collection))
      (inc! run-bytes-moved-acc (get 'bytes-moved))
      (inc! run-bytes-marked-full-acc (get 'bytes-marked-full-gc))
      (inc! run-remset-entries-added-acc 
            (apply + (get 'remset-add-transactions)))
      (inc! run-remset-entries-scanned-acc 
            (apply + (get 'remset-entries-scanned)))
      (inc! run-remset-entries-removed-acc 
            (apply + (get 'remset-entries-removed)))
      (inc! run-remset-bytes-scanned-acc 
            (apply + (get 'remset-bytes-scanned)))
      (inc! run-frames-flushed-acc (get 'frames-flushed))
      (inc! run-frames-restored-acc (get 'frames-restored))

      ; End of do-one-iter
      )

    ; Body of do-one-run

    (if (not (run.ps run))
        (set! some-failed #t)
        (set! max-process (max max-process (* 1024 (run.ps run)))))
    (let* ((iterations (run.stats run))
           (scancosts
            (let ((x (run.remset-profiles run)))
              (if (null? x)
                  (make-list (length iterations)
                             '(0 0))
                  x))))
      (for-each do-one-iter iterations scancosts)
      (set! number-of-iterations iteration) ; is same for every run
      (inc! run-count)
      (inc! run-bytes-allocated-sum run-bytes-allocated-acc)
      (inc! run-bytes-marked-sum run-bytes-marked-acc)
      (inc! run-garbage-collections-sum run-garbage-collections-acc)
      (inc! run-cpu-all-sum run-cpu-acc)
      (inc! run-cpu-gc-all-sum run-cpu-gc-acc)
      (inc! run-cpu-gc-promote-all-sum run-cpu-gc-promote-acc)
      (inc! run-cpu-gc-collect-all-sum run-cpu-gc-collect-acc)
      (inc! run-cpu-gc-full-all-sum run-cpu-gc-full-acc)
      (inc! run-full-collections-all-sum run-full-collections-acc)
      (if (< run-cpu-acc run-cpu-best)
          (let ((n (exact->inexact iteration)))
            (set! run-remset-trace-best (run.remset-trace run))
            (set! run-bytes-allocated-best run-bytes-allocated-acc)
            (set! run-garbage-collections-best run-garbage-collections-acc)
            (set! run-cpu-best run-cpu-acc)
            (set! run-cpu-user-best run-cpu-user-acc)
            (set! run-cpu-system-best run-cpu-system-acc)
            (set! run-cpu-gc-best run-cpu-gc-acc)
            (set! run-cpu-gc-promote-best run-cpu-gc-promote-acc)
            (set! run-cpu-gc-collect-best run-cpu-gc-collect-acc)
            (set! run-cpu-gc-minor-best run-cpu-gc-minor-acc)
            (set! run-cpu-gc-major-best run-cpu-gc-major-acc)
            (set! run-cpu-gc-full-best run-cpu-gc-full-acc)
            (set! run-minor-collections-best run-minor-collections-acc)
            (set! run-major-collections-best run-major-collections-acc)
            (set! run-full-collections-best run-full-collections-acc)
            (set! run-bytes-marked-best run-bytes-marked-acc)
            (set! run-bytes-copied-promotion-best
                  run-bytes-copied-promotion-acc)
            (set! run-bytes-copied-collection-best
                  run-bytes-copied-collection-acc)
            (set! run-bytes-moved-promotion-best
                  run-bytes-moved-promotion-acc)
            (set! run-bytes-moved-collection-best
                  run-bytes-moved-collection-acc)
            (set! run-bytes-copied-best run-bytes-copied-acc)
            (set! run-bytes-moved-best run-bytes-moved-acc)
            (set! run-bytes-marked-full-best run-bytes-marked-full-acc)
            (set! run-remset-entries-added-best run-remset-entries-added-acc)
            (set! run-remset-entries-scanned-best 
                  run-remset-entries-scanned-acc)
            (set! run-remset-entries-removed-best
                  run-remset-entries-removed-acc)
            (set! run-remset-bytes-scanned-best run-remset-bytes-scanned-acc)
            (set! run-frames-flushed-best run-frames-flushed-acc)
            (set! run-frames-restored-best run-frames-restored-acc)

            (set! bytes-allocated-best (/ run-bytes-allocated-best n))
            (set! garbage-collections-best (/ run-garbage-collections-best n))
            (set! cpu-best (/ run-cpu-best n))
            (set! cpu-user-best (/ run-cpu-user-best n))
            (set! cpu-system-best (/ run-cpu-system-best n))
            (set! cpu-gc-best (/ run-cpu-gc-best n))
            (set! cpu-gc-promote-best (/ run-cpu-gc-promote-best n))
            (set! cpu-gc-collect-best (/ run-cpu-gc-collect-best n))
            (set! cpu-gc-minor-best (/ run-cpu-gc-minor-best n))
            (set! cpu-gc-major-best (/ run-cpu-gc-major-best n))
            (set! cpu-gc-full-best (/ run-cpu-gc-full-best n))
            (set! minor-collections-best (/ run-minor-collections-best n))
            (set! major-collections-best (/ run-major-collections-best n))
            (set! full-collections-best (/ run-full-collections-best n))
            (set! bytes-copied-best (/ run-bytes-copied-best n))
            (set! bytes-marked-best (/ run-bytes-marked-best n))
            (set! bytes-copied-promotion-best
                  (/ run-bytes-copied-promotion-best n))
            (set! bytes-copied-collection-best
                  (/ run-bytes-copied-collection-best n))
            (set! bytes-moved-promotion-best
                  (/ run-bytes-moved-promotion-best n))
            (set! bytes-moved-collection-best
                  (/ run-bytes-moved-collection-best n))
            (set! bytes-moved-best (/ run-bytes-moved-best n))
            (set! bytes-marked-full-best (/ run-bytes-marked-full-best n))
            (set! remset-entries-added-best 
                  (/ run-remset-entries-added-best n))
            (set! remset-entries-scanned-best
                  (/ run-remset-entries-scanned-best n))
            (set! remset-entries-removed-best
                  (/ run-remset-entries-removed-best n))
            (set! remset-bytes-scanned-best 
                  (/ run-remset-bytes-scanned-best n))
            ))))

  ; This procedure accumulates the information for all iterations of
  ; all runs of a single system and benchmark.

  (define (do-one-system system)
    (if (and system-info (not (equal? system-info (cadr system))))
        (error "Failed invariant.")
        (begin
          (set! system-info (cadr system))
          (set! system-settings (cadddr system))))
    (set! benchmark-name (car system))
    (for-each do-one-run (caddr system)))

  (define (do-system-list same-systems)
    (init-iteration-accumulators)
    (init-run-accumulators)

    (for-each do-one-system same-systems)

    (compute-averages))

  ; Body of basic-statistics

  (map do-system-list (group-all-runs-by-system list-of-systems)))

; This returns a list of lists of lists where each subsublist is as
; described for generate-list-sorted-by-benchmark: (bencmark system runs
; settings) but where benchmark and system are the same for all members
; of each sublist.  (This does not take setting into account, make sure
; to filter first on setting if this makes any difference to you.)
;
; Thus the output of this procedure groups multiple runs but does not
; intermingle them, and the iterations of a single run are grouped and
; can be processed together.

(define (group-all-runs-by-system list-of-systems)
  (accumulate (lambda (a b)
                (and (eq? (car a) (car b))      ; benchmark
                     (equal? (cadr a) (cadr b)) ; system
                     ))
              (generate-list-sorted-by-benchmark list-of-systems 'trigger)))


; Sort the output by benchmark and system-name
;
; Return a list of lists: each sublist is (benchmark system runs
; settings) where benchmark is a benchmark-name, system is the <system>
; tag from the input (thus a list), and runs is a list of runs,
; unprocessed.  The list is sorted on the two first fields where both
; fields sort lexicographically, except that by passing the symbol
; 'trigger as the last argument sorts the trigger size before the heap
; size for fixed-size systems.  No merging takes place, so if there are
; several runs of the same benchmark on the same system then there will
; be multiple (consecutive) entries.

(define (generate-list-sorted-by-benchmark list-of-systems . rest)

  (define trigger-before-size? (memq 'trigger rest))

  (define (symbol<? a b)
    (string<? (symbol->string a) (symbol->string b)))

  (define (<? a b)
    (string<? (cond ((string? a) a)
                    ((symbol? a) (symbol->string a))
                    ((number? a) (number->string a))
                    (else ???))
              (cond ((string? b) b)
                    ((symbol? b) (symbol->string b))
                    ((number? b) (number->string b))
                    (else ???))))

  (define (system-name<? a b)
    (cond ((and (symbol? a) (symbol? b)) (symbol<? a b))
          ((symbol? a) (symbol<? a (car b)))
          ((symbol? b) (symbol<? (car a) b))
          (else
           (let loop ((a a) (b b))
             (cond ((null? a) #f)
                   ((<? (car a) (car b)) #t)
                   ((eq? (car a) (car b)) (loop (cdr a) (cdr b)))
                   (else #f))))))

  (define (system<? a b)
    (or (system-name<? (car a) (car b))
        (and (equal? (car a) (car b))
             (or (symbol<? (cadr a) (cadr b))
                 (and (eq? (cadr a) (cadr b))
                      (case (cadr a)
                        ((fixed upper upper+lower)
                         (if trigger-before-size?
                             (or (< (cadddr a) (cadddr b))
                                 (and (= (cadddr a) (cadddr b))
                                      (< (caddr a) (caddr b))))
                             (or (< (caddr a) (caddr b))
                                 (and (= (caddr a) (caddr b))
                                      (< (cadddr a) (cadddr b))))))
                        ((variable)
                         (< (caddr a) (caddr b)))
                        (else ???)))))))

  (define (less? a b)
    (or (symbol<? (car a) (car b))
        (and (eq? (car a) (car b))
             (system<? (cadr a) (cadr b)))))

  (sort (filter (lambda (x) x)
                (map (lambda (sys)
                       (let ((runs (system.runs sys)))
                         (if (null? runs)
                             #f
                             (list (car (run.run-number (car runs)))
                                   (system.info sys)
                                   runs
                                   (system.settings sys)))))
                     list-of-systems))
        less?))

; eof
