; Program to hack gnuplot output:
;  - replace second run of \Diamond by \diamond
;  - replace big filled circles by less obtrusive symbols

(require 'string)
(require 'unix)

(define looking-for 
  `(("\\Diamond"     . "\\Diamond")
    ("\\Diamond"     . "\\diamond")
    ("\\circle*{18}" . "$\\triangleleft$")
    ("\\circle*{24}" . "$\\triangleright$")))

(define (hack-space/time)
  (for-each (lambda (f)
              (let ((newfn
                     (string-append (substring f 0 (- (string-length f) 3))
                                    "safe")))
                (if (not (file-exists? newfn))
                    (rename-file f newfn))
                (display f)
                (newline)
                (hack-gnuplot newfn f)))
            (glob "build/plots/*-vs-*-plots/spacetime*.tex")))

(define (hack-mc-vs-L)
  (for-each (lambda (f)
              (let ((newfn
                     (string-append (substring f 0 (- (string-length f) 3))
                                    "safe")))
                (if (not (file-exists? newfn))
                    (rename-file f newfn))
                (display f)
                (newline)
                (hack-gnuplot newfn f)))
            (glob "build/plots/*-summary-plots/*.tex")))

(define (hack-gnuplot infn outfn)
  (let ((looking-for looking-for))
    (call-with-input-file infn
      (lambda (in)
        (call-with-output-file outfn
          (lambda (out)
            (call-with-current-continuation
             (lambda (k)

               (define (substring-match-assoc x)
                 (let loop ((looking-for looking-for))
                   (cond ((null? looking-for) #f)
                         ((substring-match x (caar looking-for))
                          (car looking-for))
                         (else
                          (loop (cdr looking-for))))))

               (define (get)
                 (let ((x (read-line in)))
                   (if (eof-object? x)
                       (k #t)
                       x)))

               (define (put x entry)
                 (if entry
                     (display (substring-match-replace 
                               x (car entry) (cdr entry))
                              out)
                     (display x out))
                 (newline out))

               (define (before-run x)
                 (cond ((substring-match-assoc x)
                        => (lambda (entry)
                             (begin (put x entry)
                                    (in-run (get) entry))))
                       (else
                        (put x #f)
                        (before-run (get)))))

               (define (in-run x entry)
                 (if (substring-match x (car entry))
                     (begin (put x entry) 
                            (in-run (get) entry))
                     (begin (put x #f) 
                            (set! looking-for (remq entry looking-for))
                            (before-run (get)))))

               (before-run (get))))))))))

; eof
