; Remembered set trace processing   -*- mode: scheme; mode: font-lock -*-
; 2000-08-01 / lth

(require 'format)

; Generic processing code.
;  GET is a generator -- it returns one of
;    (start)
;    (promote <remset-sizes>)
;    (collect <remset-sizes>)
;    (full <remset-sizes>)
;    (stop <remset-sizes>)
;    (end)
;  where remset-sizes is just an integer, the sum of remembered
;  set sizes following that collection.  Entire set consists of 
;  sets of (start) ... (stop) with a final (end).
;
;  The other parameters are output ports: normal data, GC marks, and 
;  full GC marks.

(define (process-rtrace get out gcout fullout)

  (define gc-times '())
  (define full-gc-times '())
  (define max-value 0)

  (define (show t n what)
    (set! max-value (max max-value n))
    (format out "~a ~a~%" t n)
    (case what
      ((C) (set! gc-times (cons (cons t n) gc-times)))
      ((F) (set! full-gc-times (cons (cons t n) full-gc-times)))
      (else #t)))

;  (define (write-time-data out data)
;    (let ((y1 (round (/ max-value 5)))
;          (y2 (round (/ max-value 10))))
;      (for-each (lambda (x)
;                  (let ((t (car x))
;                        (n (cdr x)))
;                    (format out "~a ~a~%" t (min max-value (+ n y1)))
;                    (format out "~a ~a~%~%" t (+ max-value y2))))
;                data)))

  (define (write-time-data out data adjust?)
    (let ((k (if adjust?
                 (round (* 1.15 max-value))
                 (round (* 1.2 max-value)))))
      (for-each (lambda (x)
                  (format out "~a ~a~%" (car x) k))
                data)))

  (define (before-start x t)
    (case (car x)
      ((start) (running (get) (+ t 1)))
      ((end) (finish t))
      (else (before-start (get) (+ t 1)))))

  (define (running x t)
    (case (car x)
      ((promote)
       (show t (cadr x) 'P)
       (running (get) (+ t 1)))
      ((collect) 
       (show t (cadr x) 'C)
       (running (get) (+ t 1)))
      ((full) 
       (show t (cadr x) 'F)
       (running (get) (+ t 1)))
      ((stop)
       (before-start (get) (+ t 1)))
      ((end) 
       (error "END before STOP."))
      (else
       (error "Bad tag: " x))))
      
  (define (finish t)
    (write-time-data gcout (reverse gc-times) #f)
    (write-time-data fullout (reverse full-gc-times) #t))

  (before-start (get) 0))


; DOF remembered set trace processing.
;
; This is an older trace format, used for DOF collector on February 
; 2000 traces.  Not quite obsolete, but also not super reliable.  This is 
; old code recast to use new processing code -- untested 2000-07-09, 
; should be OK.

; (process-remembered-set-trace input-file [ output-file-prefix ])

; <trace> ::= <elt>*
;
; <elt> ::= (start)
;         | (promote <n1> <n2> . <rs-data>)
;         | (collect <n1> <n2> <n3> . <rs-data>)
;         | (full <n1> . <rs-data>)
;         | (reset . <rs-data>)
;         | (stop <string>)
;
; <rs-data> ::= (<rs> ...)
;
; <rs> ::= (out <n> ...)
;        | (in <n> ...) 
;        | (sha <n> ...)
;        | (gen <n> ...)
;
; For promotion, <n1> is the chunk index of the allocation pointer before
; the collection, <n2> is the chunk index of the allocation pointer after.
;
; For collection, <n1> is the collection pointer before, <n2> is the
; reserve pointer before, and <n3> is the reserve pointer after.
;
; For full GCs, <n1> is the number of objects removed from the remembered
; sets by the full GC.
; 
; In <rs-data>, the <n> are the number of elements in the remembered sets,
; where "out" are objects in a set because they contain pointers out of
; a generation, "in" are objects in a set because they contain pointers
; into a generation, "sha" is the size of the shadow remembered sets, and
; "gen" is the size of the standard assignment set.  "in" and "out" pertain
; to the DOF remembered set.

(define (process-dof-remembered-set-trace filename . rest)

  (define (sum-remset-sizes r)
    (if (null? r)
        0
        (case (caar r)
          ((in) (sum-remset-sizes (cdr r)))
          ((out sha gen)
           (+ (sum (cdar r)) (sum-remset-sizes (cdr r))))
          (else 
           (error "sum-remset-sizes: eh? " r)))))

  (define (process in dout gcout fullout)
    (process-rtrace
     (lambda ()
       (let loop ((x (read in)))
         (if (eof-object? x) 
             '(end)
             (case (car x)
               ((promote) 
                (list 'promote (sum-remset-sizes (cdddr x))))
               ((collect) 
                (list 'collect (sum-remset-sizes (cddddr x))))
               ((full)
                (list 'full (sum-remset-sizes (cddr x))))
               ((expand reset) 
                (loop (read in)))
               (else x)))))
     dout gcout fullout))

  (format #t "Processing ~a~%" filename)
  (call-with-input-file filename
    (lambda (in)
      (if (null? rest)
          (process in (current-output-port) 
                   (current-output-port) 
                   (current-output-port))
          (call-with-output-file (string-append (car rest) ".data")
            (lambda (dout)
              (call-with-output-file (string-append (car rest) ".gc")
                (lambda (gcout)
                  (call-with-output-file (string-append (car rest) ".fgc")
                    (lambda (fullout)
                      (process in dout gcout fullout)))))))))))


; eof
