; ROF processing code          -*- mode: scheme; mode: font-lock -*-
; 2000-09-11 / lth

(require 'list)
(require 'generator)

;----------------------------------------------------------------------

(define (canned-rof-tables-and-plots)
  (display "ROF tables and plots")
  (newline)

  (load-rof-results)

  ; HTML output
  (create-rof-summary-tables 'html)
  (create-rof-summary-tables 'html 'unreduced)
  (merge-rof-summary-html-tables)
  (merge-rof-unreduced-summary-html-tables)
  (create-rof-gc-tables 'html)
  (create-rof-gc-tables 'html 'unreduced)
  (merge-rof-gc-html-tables)
  (merge-rof-unreduced-gc-html-tables)

  ; LaTeX output
  (create-rof-summary-tables 'latex)
  (create-rof-gc-tables 'latex)

  ; GNUplot output
  ; Nothing yet!

  'canned-rof-tables-and-plots)

;----------------------------------------------------------------------

(define (crunch-rof-results)
  (for-each 
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-06-01-rof.upper+lower.1" "crunched/rof.upper+lower.1")
     ("test/output/2000-06-01-rof.upper+lower.5" "crunched/rof.upper+lower.5")
     ("test/output/2000-06-01-rof.upper.1" "crunched/rof.upper.1")
     ("test/output/2000-06-01-rof.upper.5" "crunched/rof.upper.5")
     ("test/output/2000-06-01-rof.var.1" "crunched/rof.var.1")
     ("test/output/2000-06-01-rof.var.5" "crunched/rof.var.5"))))

(define (crunch-rof-float-results)
  (for-each
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-06-29-float1-rof.upper+lower.1"
      "crunched/float1-rof.upper+lower.1")
     ("test/output/2000-06-29-float1-rof.upper+lower.5"
      "crunched/float1-rof.upper+lower.5")
     ("test/output/2000-06-29-float2-rof.upper+lower.1"
      "crunched/float2-rof.upper+lower.1")
     ("test/output/2000-06-29-float2-rof.upper+lower.5"
      "crunched/float2-rof.upper+lower.5"))))

(define (crunch-rof-promotion-results)
  (for-each
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-06-29-promvol-rof.var.1" 
      "crunched/promvol-rof.var.1")
     ("test/output/2000-06-29-promvol-rof.var.5" 
      "crunched/promvol-rof.var.5"))))

(define (crunch-rof-rtrace-results)
  (for-each
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-07-06-rtrace-rof.var.1" 
      "crunched/rtrace-rof.var.1")
     ("test/output/2000-07-06-rtrace-rof.var.5" 
      "crunched/rtrace-rof.var.5"))))


(define (crunch-rof-phase-results)
  (for-each
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-07-10-phase-rof.var.1" 
      "crunched/phase-rof.var.1")
     ("test/output/2000-07-10-phase-rof.var.5" 
      "crunched/phase-rof.var.5"))))

(define (crunch-rof-extra-results)
  (for-each 
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(; EARLEY data
     ("test/output/2000-08-21-earley-rof.upper+lower.5" 
      "crunched/rof-extra.upper+lower.5")
     ("test/output/2000-08-21-earley-rof.upper.5" 
      "crunched/rof-extra.upper.5")

     ; PERM and GCOLD (note: GCOLD with "buggy" RNG)
     ("test/output/2000-08-26-rof.upper.1"
      "crunched/rof-extra2.upper.1")
     ("test/output/2000-08-26-rof.upper+lower.1"
      "crunched/rof-extra2.upper+lower.1")
     ("test/output/2000-08-26-rof.var.1"
      "crunched/rof-extra2.var.1")

     ; Phase detection data (because phase detection wasn't working)
     ("test/output/2000-08-22-rof.upper+lower.1" 
      "crunched/phase-rof-extra.upper+lower.1")
     ("test/output/2000-08-22-rof.upper+lower.5" 
      "crunched/phase-rof-extra.upper+lower.5")
     ("test/output/2000-08-22-rof.upper.1" "crunched/phase-rof-extra.upper.1")
     ("test/output/2000-08-22-rof.upper.5" "crunched/phase-rof-extra.upper.5")
     ("test/output/2000-08-22-rof.var.1" "crunched/phase-rof-extra.var.1")
     ("test/output/2000-08-22-rof.var.5" "crunched/phase-rof-extra.var.5")
     
     ; Float reduction runs -- for EARLEY and 5NBOYER3
     ("test/output/2000-08-23-float1-rof.upper+lower.5" 
      "crunched/float1-rof-extra.upper+lower.5")
     ("test/output/2000-08-23-float2-rof.upper+lower.5" 
      "crunched/float2-rof-extra.upper+lower.5")

     ; Float reduction runs for PERM, GCOLD
     ("test/output/2000-09-02-float1-rof.upper+lower.1" 
      "crunched/float1-rof-extra.upper+lower.1")
     ("test/output/2000-09-02-float2-rof.upper+lower.1" 
      "crunched/float2-rof-extra.upper+lower.1")

     ; Float reduction runs with phase detection for GCBENCH, 
     ; NBOYER, SBOYER, and TWOBIT.
     ("test/output/2000-09-04-float1-rof.upper+lower.1" 
      "crunched/float1-rof-extra2.upper+lower.1")
     ("test/output/2000-09-04-float2-rof.upper+lower.1" 
      "crunched/float2-rof-extra2.upper+lower.1")
     ("test/output/2000-09-04-float1-rof.upper+lower.5" 
      "crunched/float1-rof-extra2.upper+lower.5")
     ("test/output/2000-09-04-float2-rof.upper+lower.5" 
      "crunched/float2-rof-extra2.upper+lower.5")

     ; Promotion volume data for PERM and GCOLD
     ("test/output/2000-09-04-promvol-rof.var.1"
      "crunched/promvol-rof-extra.var.1")

     ; Promotion volume data with phase detection (actually all settings)
     ; for EARLEY, GCBENCH, NBOYER, SBOYER, and TWOBIT.
     ("test/output/2000-09-05-promvol-rof.var.1"
      "crunched/promvol-rof-extra2.var.1")
     ("test/output/2000-09-05-promvol-rof.var.5"
      "crunched/promvol-rof-extra2.var.5")

     ("test/output/2000-09-04-rtrace-rof.var.1"
      "crunched/rtrace-rof-extra.var.1")

     )))

;---------------------------------------------------------------------------

(define rof-results-loaded #f)

(define (load-rof-results)

  (if (not rof-results-loaded)
      (begin
        (set! rof-upper+lower-1 
              (load-crunched-data "crunched/rof.upper+lower.1"))
        (set! rof-upper+lower-5 
              (load-crunched-data "crunched/rof.upper+lower.5"))
        (set! rof-upper-1       (load-crunched-data "crunched/rof.upper.1"))
        (set! rof-upper-5       (load-crunched-data "crunched/rof.upper.5"))
        (set! rof-var-1         (load-crunched-data "crunched/rof.var.1"))
        (set! rof-var-5         (load-crunched-data "crunched/rof.var.5"))

        ; Merge in EARLEY results
        (set! rof-upper+lower-5
              (append (remove-benchmark '(5 earley:12) rof-upper+lower-5)
                      (load-crunched-data "crunched/rof-extra.upper+lower.5")))
        (set! rof-upper-5
              (append (remove-benchmark '(5 earley:12) rof-upper-5)
                      (load-crunched-data "crunched/rof-extra.upper.5")))

        ; Merge in the new phase detection results, and manipulate
        ; option settings on existing data to correspond to the actual
        ; situation: any data on old runs that supposedly had phase
        ; detection on, did not.  
        ; The data we're merging in has two switch settings: phase
        ; detection alone, and all three tuning parameters together.
        (set! rof-upper+lower-1
              (append (remove-phase-detection rof-upper+lower-1)
                      (load-crunched-data 
                       "crunched/phase-rof-extra.upper+lower.1")))
        (set! rof-upper+lower-5
              (append (remove-phase-detection rof-upper+lower-5)
                      (load-crunched-data 
                       "crunched/phase-rof-extra.upper+lower.5")))
        (set! rof-upper-1
              (append (remove-phase-detection rof-upper-1)
                      (load-crunched-data "crunched/phase-rof-extra.upper.1")))
        (set! rof-upper-5
              (append (remove-phase-detection rof-upper-5)
                      (load-crunched-data "crunched/phase-rof-extra.upper.5")))
        (set! rof-var-1
              (append (remove-phase-detection rof-var-1)
                      (load-crunched-data "crunched/phase-rof-extra.var.1")))
        (set! rof-var-5
              (append (remove-phase-detection rof-var-5)
                      (load-crunched-data "crunched/phase-rof-extra.var.5")))

        ; Merge in PERM and GCOLD results.
        (set! rof-upper-1
              (append rof-upper-1
                      (load-crunched-data "crunched/rof-extra2.upper.1")))
        (set! rof-upper+lower-1
              (append rof-upper+lower-1
                      (load-crunched-data 
                       "crunched/rof-extra2.upper+lower.1")))
        (set! rof-var-1
              (append rof-var-1
                      (load-crunched-data "crunched/rof-extra2.var.1")))

        (set! rof-results-loaded #t)
        (unspecified))))


(define rof-float-results-loaded #f)

(define (load-rof-float-results)
  (if (not rof-float-results-loaded)
      (begin
        (set! float-rof-upper+lower-1
              (append 
               (crunched-add-option! (load-crunched-data
                                      "crunched/float1-rof.upper+lower.1")
                                     '-rof-fullgc-on-collection)
               (crunched-add-option! (load-crunched-data 
                                      "crunched/float2-rof.upper+lower.1")
                                     '-rof-fullgc-on-promotion)))
        (set! float-rof-upper+lower-5
              (append 
               (crunched-add-option! (load-crunched-data 
                                      "crunched/float1-rof.upper+lower.5")
                                     '-rof-fullgc-on-collection)
               (crunched-add-option! (load-crunched-data 
                                      "crunched/float2-rof.upper+lower.5")
                                     '-rof-fullgc-on-promotion)))

        ; Wash the data to deal with the phase detection issue.
        (set! float-rof-upper+lower-1
              (remove-phase-detection float-rof-upper+lower-1))
        (set! float-rof-upper+lower-5
              (remove-phase-detection float-rof-upper+lower-5))

        ; Float catchup runs for EARLEY; data set also contains
        ; 5NBOYER3 but that was ineffective, so remove it.
        (set! float-rof-upper+lower-5
              (append (remove-benchmarks '((5 nboyer:3)
                                           (5 earley:12))
                                         float-rof-upper+lower-5)
                      (crunched-add-option! 
                       (remove-benchmark '(5 nboyer:3)
                                         (load-crunched-data 
                                          "crunched/float1-rof-extra.upper+lower.5"))
                       '-rof-fullgc-on-collection)
                      (crunched-add-option!
                       (remove-benchmark '(5 nboyer:3)
                                         (load-crunched-data
                                          "crunched/float2-rof-extra.upper+lower.5"))
                       '-rof-fullgc-on-promotion)))

        ; Float runs for PERM and GCOLD
        (set! float-rof-upper+lower-1
              (append float-rof-upper+lower-1
                      (crunched-add-option!
                       (load-crunched-data
                        "crunched/float1-rof-extra.upper+lower.1")
                       '-rof-fullgc-on-collection)
                      (crunched-add-option!
                       (load-crunched-data
                        "crunched/float2-rof-extra.upper+lower.1")
                       '-rof-fullgc-on-promotion)))

        ; Float runs with correct phase data for GCBENCH, NBOYER,
        ; SBOYER, and TWOBIT.
        (set! float-rof-upper+lower-1
              (append float-rof-upper+lower-1
                      (crunched-add-option!
                       (load-crunched-data
                        "crunched/float1-rof-extra2.upper+lower.1")
                       '-rof-fullgc-on-collection)
                      (crunched-add-option!
                       (load-crunched-data
                        "crunched/float2-rof-extra2.upper+lower.1")
                       '-rof-fullgc-on-promotion)))
        (set! float-rof-upper+lower-5
              (append float-rof-upper+lower-5
                      (crunched-add-option!
                       (load-crunched-data
                        "crunched/float1-rof-extra2.upper+lower.5")
                       '-rof-fullgc-on-collection)
                      (crunched-add-option!
                       (load-crunched-data
                        "crunched/float2-rof-extra2.upper+lower.5")
                       '-rof-fullgc-on-promotion)))

        (set! rof-float-results-loaded #t)
        (unspecified))))


(define rof-promvol-results-loaded #f)

(define (load-rof-promvol-results)
  (if (not rof-promvol-results-loaded)
      (begin
        ; Old data
;        (set! promvol-rof-var-1 
;              (load-crunched-data "crunched/promvol-rof.var.1"))
;        (set! promvol-rof-var-5
;              (load-crunched-data "crunched/promvol-rof.var.5"))

        ; New data
        (set! promvol-rof-var-1 
              (append
               (load-crunched-data "crunched/promvol-rof-extra.var.1")
               (load-crunched-data "crunched/promvol-rof-extra2.var.1")))
        (set! promvol-rof-var-5
              (load-crunched-data "crunched/promvol-rof-extra2.var.5"))

        (set! rof-promvol-results-loaded #t)
        (unspecified))))


(define rof-rtrace-results-loaded #f)

(define (load-rof-rtrace-results)
  (if (not rof-rtrace-results-loaded)
      (begin
        (set! rtrace-rof-var-1 
              (load-crunched-data "crunched/rtrace-rof.var.1"))
        (set! rtrace-rof-var-5 
              (load-crunched-data "crunched/rtrace-rof.var.5"))

        ; PERM, GCOLD
        (set! rtrace-rof-var-1
              (append rtrace-rof-var-1
                      (load-crunched-data "crunched/rtrace-rof-extra.var.1")))

        (set! rof-rtrace-results-loaded #t)
        (unspecified))))


(define rof-phase-results-loaded #f)

(define (load-rof-phase-results)
  (if (not rof-phase-results-loaded)
      (begin
        (set! phase-rof-var-1 (load-crunched-data "crunched/phase-rof.var.1"))
        (set! phase-rof-var-5 (load-crunched-data "crunched/phase-rof.var.5"))
        (set! rof-phase-results-loaded #t)
        (unspecified))))


; Only removes the first one.

(define (remove-subsequence same? s l)
  (cond ((null? s)
         l)
        ((null? l)
         '())
        ((same? (car s) (car l))
         (remove-subsequence same? (cdr s) (cdr l)))
        (else
         (cons (car l) (remove-subsequence same? s (cdr l))))))

; Wash the data: if a run has only phase detection set, then remove
; it altogether, and otherwise remove the phase detection switch.

(define (remove-phase-detection datum)
  (map (lambda (system)
         (system.info! system
                       (remove-subsequence eqv? 
                                           '(-phase-detection 0.001)
                                           (system.info system))))
       (filter (lambda (system)
                 (let ((i (system.info system)))
                   (not (and (memq '-phase-detection i)
                             (not (memq '-feeling-lucky i))
                             (not (memq '-np-remset-limit i))))))
               datum)))

; Select the data set with the canonical settings.

; Five runs:
;    -phase-detection 0.001    -feeling-lucky 0.5    -np-remset-limit 32K
; 1           0                        0                     0
; 2           1                        0                     0
; 3           1                        1                     0
; 4           1                        0                     1
; 5           1                        1                     1
;
; I've selected the last run as the canonical run, because this usually
; has the best performance.  However, those settings also cause the
; collector to fail more often from memory exhaustion in fixed heaps.

(define (rof-canonical-setting? datum)  ; datum from BASIC-STATISTICS

  (define (present? tag)
    (assq tag datum))

  (and (present? '-phase-detection)
       (present? '-feeling-lucky)
       (present? '-np-remset-limit)))
  
(define (rof-extract-canonical data)
  (select-most-desirable-run-and-resort 
   data
   (lambda (a b)
     (and (rof-canonical-setting? a) 
          (not (rof-canonical-setting? b))))))

(define (rof-compute-canonical-data raw-data)
  (rof-extract-canonical 
   (flatten-option-noise 
    (fluid-let ((no-warnings #t))
      (basic-statistics raw-data)))))

(define (rof-compute-float-data raw-data selectors)
  (let ((non-selectors (filter (lambda (x)
                                 (not (memq x selectors)))
                               '(-phase-detection 
                                 -feeling-lucky 
                                 -np-remset-limit))))

    (filter (lambda (a)
              (and (every? (lambda (x) (assq x a)) selectors)
                   (every? (lambda (x) (not (assq x a))) non-selectors)))
            (select-trigger
             3.0
             (flatten-option-noise 
              (fluid-let ((no-warnings #t))
                (basic-statistics raw-data)))))))

'(define (test-float-selector selectors)
  (map (lambda (x)
         (list (benchmark-designator x)
               (list-head (assq 'system-info x) 5)
               (assq '-phase-detection x)
               (assq '-feeling-lucky x)
               (assq '-np-remset-limit x)))
       (rof-compute-float-data rof-upper+lower-1 selectors)))

(define (rof-canonical-data)
  (load-rof-results)
  `((,(rof-compute-canonical-data rof-upper-1) 1)
    (,(rof-compute-canonical-data rof-upper-5) 5)
    (,(rof-compute-canonical-data rof-upper+lower-1) 1)
    (,(rof-compute-canonical-data rof-upper+lower-5) 5)
    (,(rof-compute-canonical-data rof-var-1) 1)
    (,(rof-compute-canonical-data rof-var-5) 5)))

(define (rof-float-data selectors)
  (load-rof-float-results)
  (load-rof-results)
  `((,(rof-compute-float-data rof-upper+lower-1 selectors) 1)
    (,(rof-compute-float-data rof-upper+lower-5 selectors) 5)
    (,(rof-compute-float-data float-rof-upper+lower-1 selectors) 1)
    (,(rof-compute-float-data float-rof-upper+lower-5 selectors) 5)))

(define (rof-noncanonical-results . selectors)
  (let ((non-selectors (filter (lambda (x)
                                 (not (memq x selectors)))
                               '(-phase-detection 
                                 -feeling-lucky 
                                 -np-remset-limit))))

    (define (compute raw-data)
      (select-most-desirable-run-and-resort
       (flatten-option-noise
        (fluid-let ((no-warnings #t))
          (basic-statistics raw-data)))
       (lambda (a b)
         (and (every? (lambda (x) (assq x a)) selectors)
              (every? (lambda (x) (not (assq x a))) non-selectors)))))
    
    `((,(compute rof-upper-1) 1)
      (,(compute rof-upper-5) 5)
      (,(compute rof-upper+lower-1) 1)
      (,(compute rof-upper+lower-5) 5)
      (,(compute rof-var-1) 1)
      (,(compute rof-var-5) 5))))

(define (generic-rof-summary-tables prefix suffix variables stats1 stats5 type)
  (do ((variables variables (cdr variables)))
      ((null? variables))
    (let* ((var  (car variables))
           (s    (symbol->string var))
           (file (case type
                   ((latex) (string-append "tables/" prefix s suffix ".tex"))
                   ((html)  (string-append "html-tables/" 
                                           prefix s suffix ".html"))
                   ((text)  "/dev/tty")
                   (else ???)))
           (n    (- (char->integer 
                     (string-ref s (- (string-length s) 1)))
                    (char->integer #\0)))
           (data (environment-get (interaction-environment) 
                                  var))
           (limit (not (string=? "var" (substring s 4 7)))))
      (display var)
      (newline)
      (fluid-let ((field-width 9))
        (with-output-to-file file
          (lambda ()
            (if (= n 1)
                (if limit
                    (stats1 data 'no-load/trigger)
                    (stats1 data 'no-heap-limit))
                (if limit
                    (stats5 data 'no-load/trigger)
                    (stats5 data 'no-heap-limit)))))))))

(define (create-rof-summary-tables . rest)

  (define filter 
    (lambda (x)
      (rof-extract-canonical
       (flatten-option-noise x))))

  (define (fudge-factors get present?)
    (list (present? '-phase-detection)
          (present? '-feeling-lucky)
          (present? '-np-remset-limit)))

  (define type 'latex)
  (define unreduced #f)
  (define suffix "")

  (define (stats list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")         ms run-cpu-best)
             (("CPU GC")      ms run-cpu-gc-best)
             (("CPU" "Promote")  ms run-cpu-gc-promote-best)
             (("Major" "GCs") n  run-major-collections-best)
             (("Marked")        mb run-bytes-marked-best)
             (("Heap" "used") mb bytes-max-heap))
           'no-warnings 
           'break 9 ; workaround for bug, was: (if (eq? type 'latex) 9 10000)
           type 
           'no-variance
           'no-fail
           'filter filter
           'sort sort-by-trigger
           'header-filter (lambda (bm collector heap aux)
                            (values (format-benchmark-name bm)
                                    "ROF"
                                    heap
                                    (if (null? aux) 
                                        #f
                                        (format #f "T=~a" (car aux)))))
           (if unreduced
               (cons 'break-dof-unreduced rest)
               rest)))

  (for-each (lambda (x)
              (case x
                ((text)      (set! type 'text))
                ((latex)     (set! type 'latex))
                ((html)      (set! type 'html))
                ((unreduced) (set! unreduced #t))
                (else ???)))
            rest)

  (if unreduced
      (begin
        (set! filter flatten-option-noise)
        (set! suffix "-u")))

  (load-rof-results)
  (generic-rof-summary-tables 
   "" suffix
   '(rof-upper-1 rof-upper-5
     rof-upper+lower-1 rof-upper+lower-5
     rof-var-1 rof-var-5)
   stats stats
   type))


(define (merge-rof-summary-html-tables)
  (merge-summary-html-tables
   "html-tables/rof-general.html"
   "ROF results | General"
    '("rof-upper+lower-1" "rof-upper+lower-5"
      "rof-upper-1" "rof-upper-5"
      "rof-var-1" "rof-var-5")))

(define (merge-rof-unreduced-summary-html-tables)
  (merge-summary-html-tables
   "html-tables/rof-general-unreduced.html"
   "ROF results | General"
    '("rof-upper+lower-1-u" "rof-upper+lower-5-u"
      "rof-upper-1-u" "rof-upper-5-u"
      "rof-var-1-u" "rof-var-5-u")))


; Note, remset peak is peak _allocated_, not peak used.

(define (create-rof-gc-tables . rest)

  (define filter 
    (lambda (x)
      (rof-extract-canonical
       (flatten-option-noise x))))

  (define (fudge-factors get present?)
    (list (present? '-phase-detection)
          (present? '-feeling-lucky)
          (present? '-np-remset-limit)))

  (define type 'text)
  (define unreduced #f)
  (define suffix "")

  (define (stats1 list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")             ms cpu-best)
             (("CPU" "GC")        ms cpu-gc-best)
             (("CPU" "User")      ms cpu-user-best)
             (("CPU" "Sys.")      ms cpu-system-best)
             (("CPU" "Mut.")      ms ,(lambda (get present?)
                                        (- (get 'cpu-best) 
                                           (get 'cpu-gc-best))))
             (("Copy" "gc")       mb bytes-copied-best)
             (("Move" "los")      mb bytes-moved-best)
             (("Objs" "inserted") exact.kf.1 remset-entries-added-best)
             (("Objs" "scanned")  exact.kf.1 remset-entries-scanned-best)
             (("Bytes" "scanned") exact.kf.1 remset-bytes-scanned-best)
             (("Objs" "removed")  exact.kf.1 remset-entries-removed-best)
             (("Bytes/" "object") nf.1 
                                  ,(lambda (get present?)
                                     (exact->inexact
                                      (/ (get 'remset-bytes-scanned-best)
                                         (get 'remset-entries-scanned-best)))))
             (("Remset" "peak")   kb bytes-max-remset)
             (("ph/lu/li") bv ,fudge-factors))
           'no-warnings 
           'break (if (eq? type 'latex) 10 10000) 
           type 
           'no-variance 
           'no-load/trigger
           'filter filter
           (if unreduced
               (cons 'break-dof-unreduced rest)
               rest)))

  (define (stats5 list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")             ms run-cpu-best)
             (("CPU" "GC")        ms run-cpu-gc-best)
             (("CPU" "User")      ms run-cpu-user-best)
             (("CPU" "Sys.")      ms run-cpu-system-best)
             (("CPU" "Mut.")      ms ,(lambda (get present?)
                                        (- (get 'run-cpu-best) 
                                           (get 'run-cpu-gc-best))))
             (("Copy" "gc")       mb run-bytes-copied-best)
             (("Move" "los")      mb run-bytes-moved-best)
             (("Objs" "inserted") kf.1 run-remset-entries-added-best)
             (("Objs" "scanned")  kf.1 run-remset-entries-scanned-best)
             (("Bytes" "scanned") kf.1 run-remset-bytes-scanned-best)
             (("Objs" "removed")  kf.1 run-remset-entries-removed-best)
             (("Bytes/" "object") nf.1 
              ,(lambda (get present?)
                 (exact->inexact
                  (/ (get 'run-remset-bytes-scanned-best)
                     (get 'run-remset-entries-scanned-best)))))
             (("Remset" "peak")   kb bytes-max-remset)
             (("ph/lu/li") bv ,fudge-factors))
           'no-warnings 
           'break (if (eq? type 'latex) 10 10000) 
           type 
           'no-variance 
           'no-load/trigger
           'filter filter
           (if unreduced
               (cons 'break-dof-unreduced rest)
               rest)))

  (for-each (lambda (x)
              (case x
                ((text)      (set! type 'text))
                ((latex)     (set! type 'latex))
                ((html)      (set! type 'html))
                ((unreduced) (set! unreduced #t))
                (else ???)))
            rest)

  (if unreduced
      (begin
        (set! filter flatten-option-noise)
        (set! suffix "-u")))

  (generic-rof-summary-tables 
   "gc-" suffix
   '(rof-upper-1 rof-upper-5
     rof-upper+lower-1 rof-upper+lower-5
     rof-var-1 rof-var-5)
   stats1 stats5
   type))

(define (merge-rof-gc-html-tables)
  (merge-summary-html-tables
   "html-tables/rof-gc.html"
   "ROF results | GC"
    '("gc-rof-upper+lower-1" "gc-rof-upper+lower-5"
      "gc-rof-upper-1" "gc-rof-upper-5"
      "gc-rof-var-1" "gc-rof-var-5")))

(define (merge-rof-unreduced-gc-html-tables)
  (merge-summary-html-tables
   "html-tables/rof-gc-unreduced.html"
   "ROF results | GC"
    '("gc-rof-upper+lower-1-u" "gc-rof-upper+lower-5-u"
      "gc-rof-upper-1-u" "gc-rof-upper-5-u"
      "gc-rof-var-1-u" "gc-rof-var-5-u")))

; Table of mark/cons numbers.

(define (rof-mark/cons-analysis . rest)

  (define (process-datum datum n results)
    (if (all-iterations-succeeded? datum)
        (let* ((bm (list n (cdr (assq 'benchmark-name datum))))
               (cell ((results 'lookup) bm)))
          (set-car! (cdr cell) 
                    (+ (cadr cell)
                       (/ (+ (cdr (assq 'run-bytes-copied-best datum))
                             (cdr (assq 'run-bytes-moved-best datum))
                             (cdr (assq 'run-bytes-marked-full-best datum)))
                          (canonical-allocation bm))))
          (set-car! (cddr cell) (+ (caddr cell) 1))
          (set-car! (cdddr cell) 
                    (+ (cadddr cell)
                       (cdr (assq 'run-major-collections-best datum)))))))

  (define (analyze tag all-data)
    (let ((results
           (make-benchmark-associator (lambda () (list 0 0 0)))))
      (for-each-benchmark-result 
       (lambda (datum n) (process-datum datum n results))
       all-data)
      (map (lambda (x) 
             (list (cons tag (car x)) 
                   (/ (cadr x) (caddr x))
                   (/ (cadddr x) (caddr x))))
           ((results 'results)))))

  (define (mark/cons-analysis)
    (let ((results (append (analyze 0 (2gen-canonical-data))
                           (analyze 1 (3gen-canonical-data))
                           (analyze 2 (rof-canonical-data)))))
      (map (lambda (bm)
             `(,bm 
               ,@(map 
                  (lambda (gc)
                    (let ((v (assoc (cons gc bm) results)))
                      (if v
                          (list gc (cadr v))
                          (list gc #f))))
                  (iota 3))
               ,@(map 
                  (lambda (gc)
                    (let ((v (assoc (cons gc bm) results)))
                      (if v
                          (list gc (caddr v))
                          (list gc #f))))
                  (iota 3))))
           benchmark-names)))

  (define (display-analysis data out)
    (latex-table
     out
     "|l|rrr|rrr|"
     '(("" ("Mark/cons ratio" 3) ("Major GCs" 3))
       ("" "2GEN" "3GEN" "ROF" "2GEN" "3GEN" "ROF"))
     (cons '(benchmark-name "~20a") (make-list 6 '(num-cadr "~2,2f")))
     data))

  (if (null? rest)
      (display-analysis (mark/cons-analysis) (current-output-port))
      (call-with-output-medium (car rest) 
        (lambda (out)
          (display-analysis (mark/cons-analysis) out)))))


; Remset analysis
;
; Table of peak remset size as a percentage of peak heap size (for that
; run), averaged over all runs for a particular benchmark and collector.

(define (rof-remset-analysis . rest)

;  (define (process-datum datum iter results)
;    (if (all-iterations-succeeded? datum)
;        (let ((n 
;               (* 100.0 (/ (cdr (assq 'bytes-max-remset datum))
;                           (cdr (assq 'bytes-max-heap datum)))))
;              (cell 
;               ((results 'lookup) (benchmark-designator datum))))
;          (set-car! (cdr cell) (+ (cadr cell) n))
;          (set-car! (cddr cell) (+ (caddr cell) 1)))))

  (define megabyte (* 1024 1024))

  (define (process-datum datum iter results)
    (if (all-iterations-succeeded? datum)
        (let ((n    (/ (cdr (assq 'bytes-max-remset datum)) megabyte))
              (cell ((results 'lookup) (benchmark-designator datum))))
          (set-car! (cdr cell) (+ (cadr cell) n))
          (set-car! (cddr cell) (+ (caddr cell) 1)))))

  (define (analyze tag all-data)
    (let ((results
           (make-benchmark-associator (lambda () 
                                        (list 0 0)))))
      (for-each-benchmark-result 
       (lambda (datum iter)
         (process-datum datum iter results))
       all-data)
      (map (lambda (x) 
             (cons (cons tag (car x)) (/ (cadr x) (caddr x))))
           ((results 'results)))))

  (define (remset-analysis)
    (let ((results
           (append (analyze 0 (2gen-canonical-data))
                   (analyze 1 (3gen-canonical-data))
                   (analyze 2 (rof-canonical-data)))))
      (map (lambda (bm)
             (cons bm
                   (map
                    (lambda (gc)
                      (let ((probe (assoc (cons gc bm) results)))
                        (if probe
                            (list gc (cdr probe))
                            (list gc #f))))
                    (iota 3))))
           benchmark-names)))

  (define (display-analysis data out)
    (latex-table
     out
     "|l|rrr|"
     '(("" "2GEN" "3GEN" "ROF"))
     (cons '(benchmark-name "~20a") (make-list 3 '(num-cadr "~5,2f")))
     (sort-by-benchmark-name data car)))

  (if (null? rest)
      (display-analysis (remset-analysis) (current-output-port))
      (call-with-output-medium (car rest) 
        (lambda (out)
          (display-analysis (remset-analysis) out)))))


; Table of remset insertions, objects scanned, words scanned.

(define (rof-remset-activity-analysis . rest)

  (define (init) (make-list 5 0))

  (define (get x n) (list-ref x n))
  (define (put x n v) (list-set! x n (+ (list-ref x n) v)))

  (define (rinsert x . rest) (if (null? rest) (get x 1) (put x 1 (car rest))))
  (define (rremove x . rest) (if (null? rest) (get x 2) (put x 2 (car rest))))
  (define (rcount x . rest) (if (null? rest) (get x 3) (put x 3 (car rest))))
  (define (rscan x . rest) (if (null? rest) (get x 4) (put x 4 (car rest))))
  (define (rwscan x . rest) (if (null? rest) (get x 5) (put x 5 (car rest))))
  
  (define (process-datum datum iter results)
    (if (all-iterations-succeeded? datum)
        (let ((cell ((results 'lookup)
                     (list iter (cdr (assq 'benchmark-name datum))))))
          (rinsert cell (cdr (assq 'run-remset-entries-added-best datum)))
          (rremove cell (cdr (assq 'run-remset-entries-removed-best datum)))
          (rcount cell 1)
          (rscan cell (cdr (assq 'run-remset-entries-scanned-best datum)))
          (rwscan cell (/ (cdr (assq 'run-remset-bytes-scanned-best datum)) 
                          4)))))

  (define (analyze tag all-data)
    (let ((results (make-benchmark-associator init)))
      (for-each-benchmark-result 
       (lambda (datum iter) (process-datum datum iter results))
       all-data)
      (map (lambda (x)                    
             ; Note, compatible representation!
             (list (cons tag (car x))
                   (/ (rinsert x) (rcount x))
                   (/ (rremove x) (rcount x))
                   (rcount x)
                   (/ (rscan x) (rcount x))
                   (/ (rwscan x) (rcount x))))
           ((results 'results)))))

  (define (remset-analysis)
    (let ((results
           (append (analyze 0 (2gen-canonical-data))
                   (analyze 1 (3gen-canonical-data))
                   (analyze 2 (rof-canonical-data))
                   (analyze 3 (rof-noncanonical-results)))))
      (map (lambda (bm)
             (cons bm
                   (map
                    (lambda (gc)
                      (let ((probe (assoc (cons gc bm) results)))
                        (if probe
                            ; Note, compatible representation!
                            (list gc 
                                  (rinsert probe)
                                  (rremove probe)
                                  (rcount probe)
                                  (rscan probe)
                                  (rwscan probe))
                            (list gc #f #f #f #f #f))))
                    (iota 4))))
           benchmark-names)))

  (define (sci3 x)
    (if (not x)
        ""
        (format-byte-volume.1 (exact->inexact x))))

  (define (display-analysis name data out f)
    (latex-table
     out
     "|l|rrrr|"
     `(("" (,name 4))
       ("" "2GEN" "3GEN" "ROF" "ROF-NC"))
     (cons '(benchmark-name "~20a")
           (make-list 4 (lambda (d i) (sci3 (f (list-ref d i))))))
     (sort-by-benchmark-name data car)))

  (define (show-results results out)
    (do ((f '("Objects inserted" "Objects scanned" "Words scanned") (cdr f))
         (k '(1 4 5) (cdr k)))
        ((null? f))
      (display-analysis (car f) results out (lambda (x)
                                              (list-ref x (car k))))))

  (if (null? rest)
      (show-results (remset-analysis) (current-output-port))
      (call-with-output-medium (car rest) 
        (lambda (out)
          (show-results (remset-analysis) out)))))

; What we want is to compare a run with the feature to a run without
; the feature, so:
;
;  000 vs 100  (determines phase-detection)
;  000 vs 010  (determines feeling-lucky)
;  000 vs 001  (determines np-remset-limit)
;  000 vs 111  (determines cumulative impact)

(define (rof-feature-evaluation-table how . rest)

  (define field
    (cond ((eq? how 'cpu-time) 'run-cpu-gc-best)
          ((eq? how 'total-time) 'run-cpu-best)
          (else (error "Either CPU-TIME or TOTAL-TIME: " how))))

  (define (get x) 
    (cdr (assq field x)))

  (define (compare a b)
    (compare-by-benchmark get get (list a) (list b) identity))

  (define (feature-analysis)
    (let ((r000 (rof-noncanonical-results))
          (r100 (rof-noncanonical-results '-phase-detection))
          (r010 (rof-noncanonical-results '-feeling-lucky))
          (r001 (rof-noncanonical-results '-np-remset-limit))
          (r111 (rof-noncanonical-results '-phase-detection
                                          '-feeling-lucky
                                          '-np-remset-limit)))
      (map (lambda (bm)
             (cons bm (map (lambda (r)
                             (weighted-mean (cdr (assoc bm r)) car cdr))
                           (list (compare r000 r100)
                                 (compare r000 r010)
                                 (compare r000 r001)
                                 (compare r000 r111)))))
           benchmark-names)))

  (define (display-results results out)
    (latex-table
     out
     "|l|rrrr|"
     '(("" "Phase" "Luck" "Limit"  "All"))
     (cons '(benchmark-name "~20a") (make-list 4 '(literal "~5,2f")))
     results))

  (if (null? rest)
      (display-results (feature-analysis) (current-output-port))
      (call-with-output-medium (car rest)
        (lambda (out)
          (display-results (feature-analysis) out)))))


; Promotion volumes and costs.

(define (rof-extract-promotion/collection-volume load)
  (load-rof-promvol-results)
  (append (extract-promotion/collection-volume
           (select-load-factor load (basic-statistics promvol-rof-var-1)))
          (extract-promotion/collection-volume
           (select-load-factor load (basic-statistics promvol-rof-var-5)))))

(define (rof-extract-promotion/collection-cost load)
  (load-rof-results)
  (let ((r1 (select-load-factor load (basic-statistics rof-var-1)))
        (r2 (select-load-factor load (basic-statistics rof-var-5))))
    (map (lambda (x)
           (list (caar x) 
                 (mean (map cadr x))
                 (mean (map caddr x))))
         (accumulate-unordered
          (lambda (a b)
            (equal? (car a) (car b)))
          (filter-map (lambda (x)
                        (if (all-iterations-succeeded? x)
                            (list (benchmark-designator x)
                                  (cdr (assq 'run-cpu-gc-promote-best x))
                                  (cdr (assq 'run-cpu-gc-collect-best x)))
                            #f))
                      (append r1 r2))))))

; Computes the cost of promotion (ms/MB).

(define (rof-promotion/collection-cost load)
  (let ((v (rof-extract-promotion/collection-volume load))
        (c (rof-extract-promotion/collection-cost load)))
    (map (lambda (x)
           (let ((y (cdr (assoc (car x) v))))
             (list (car x)
                   (exact->inexact (/ (cadr x) 
                                      (/ (+ (caar y) (cdar y))
                                         (* 1024 1024))))
                   (exact->inexact (/ (caddr x)
                                      (/ (+ (caadr y) (cdadr y))
                                         (* 1024 1024)))))))
         c)))


; Float analysis

(define (all-rof-mark-volumes)

  (define (display-results tag data out)
    (latex-table 
     out
     "|l|rrrrr|"
     `(("" (,(symbol->string tag) 5))
       ("" "(none)" "Phase" "Luck"  "Limit" "Phase")
       ("" ""       ""      ""      ""      "Luck")
       ("" ""       ""      ""      ""      "Limit"))
     (cons '(benchmark-name "~20a") 
           (make-list 5 `(,(lambda (datum n)
                             (let ((x (list-ref datum n)))
                               (if (not x) "" (kf.1 (/ x (* 1024 1024))))))
                          "~8@a")))
     data))

  (for-each 
   (lambda (tag)
     (display-results 
      tag 
      (let ((x (list (rof-mark-volumes tag '())
                     (rof-mark-volumes tag '(-phase-detection))
                     (rof-mark-volumes tag '(-feeling-lucky))
                     (rof-mark-volumes tag '(-np-remset-limit))
                     (rof-mark-volumes tag '(-phase-detection
                                             -feeling-lucky
                                             -np-remset-limit)))))
        (apply map (lambda xs (cons (caar xs)
                                    (map cadr xs)))
               x))
      (current-output-port)))
   '(normal collect promote)))

; As above, but compute the volume of promotion and collection float
; for each setting.

(define (all-rof-mark-volumes-differences)

  (define (display-results tag data out)
    (latex-table 
     out
     "|l|rrrrr|"
     `(("" (,tag 5))
       ("" "(none)" "Phase" "Luck"  "Limit" "Phase")
       ("" ""       ""      ""      ""      "Luck")
       ("" ""       ""      ""      ""      "Limit"))
     (cons '(benchmark-name "~20a") 
           (make-list 5 `(,(lambda (datum n)
                             (let ((x (list-ref datum n)))
                               (if (not x) "" (kf.1 (/ x (* 1024 1024))))))
                          "~8@a")))
     data))

  (define (display-differences tag a b out)
    (display-results tag 
                     (map (lambda (a b)
                            (cons (car a)
                                  (map (lambda (x y) (if (and x y) (- x y) #f)) (cdr a) (cdr b))))
                          a b)
                     out))

  (define (compute-volume tag)
    (let ((x (list (rof-mark-volumes tag '())
                   (rof-mark-volumes tag '(-phase-detection))
                   (rof-mark-volumes tag '(-feeling-lucky))
                   (rof-mark-volumes tag '(-np-remset-limit))
                   (rof-mark-volumes tag '(-phase-detection
                                           -feeling-lucky
                                           -np-remset-limit)))))
      (apply map (lambda xs (cons (caar xs)
                                  (map cadr xs)))
             x)))

  (let ((normal (compute-volume 'normal))
        (collect (compute-volume 'collect))
        (promote (compute-volume 'promote)))
    (display-differences "collection float" normal collect (current-output-port))
    (display-differences "promotion float" collect promote (current-output-port))))


(define (rof-mark-volumes tag settings)
  (let ((data (rof-float-analysis-core (rof-float-data settings))))
    (map (lambda (bm)
           (list bm
                 (let ((x (assoc (cons bm tag) data)))
                   (if (not x)
                       #f
                       (exact->inexact (cdr x))))))
         benchmark-names)))


; Table of copy volume across benchmarks (down) and ROF collector (across),
; for both full GC before ROF GC (left half) and full GC before any GC
; (right half), expressed as a percentage reduction relative to the
; canonical run.

(define (rof-float-analysis . rest)

  (define (float-analysis)
    (let ((results (rof-float-analysis-core 
                    (rof-float-data '(-phase-detection
                                      -feeling-lucky
                                      -np-remset-limit)))))
      (map (lambda (bm)
             (cons bm (map (lambda (full)
                             (let ((normal  (assoc (cons bm 'normal) results))
                                   (reduced (assoc (cons bm full) results)))
                               (if (and normal reduced)
                                   (* 100.0 
                                      (- 1.0 (/ (cdr reduced) (cdr normal))))
                                   #f)))
                           '(collect promote))))
           benchmark-names)))

  (define (display-analysis data out)
    (latex-table 
     out 
     "|l|rr|"
     '(("" "Before ROF GC" "Before every GC"))
     '((benchmark-name "~20a") (literal "~5,1f") (literal "~5,1f"))
     data))

  (if (null? rest)
      (display-analysis (float-analysis) (current-output-port))
      (call-with-output-file (car rest) 
        (lambda (out)
          (display-analysis (float-analysis) out)))))

; This is like above, but presents derived data: collection float and
; promotion float as a percentage increase in copying over the minimum.

(define (rof-float-analysis-derived selectors)

  (define (float-analysis)
    (let ((results (rof-float-analysis-core 
                    (rof-float-data selectors))))
      (map (lambda (bm)
             (cons bm 
                   (map (lambda (full)
                          (let ((normal  (assoc (cons bm 'normal) results))
                                (collect (assoc (cons bm 'collect) results))
                                (promote (assoc (cons bm 'promote) results)))
                            (case full
                              ((collect)
                               (if (and normal collect)
                                   (* 100.0 
                                      (- (/ (cdr normal) (cdr collect)) 1.0))
                                   #f))
                              ((promote)
                               (if (and collect promote)
                                   (* 100.0 
                                      (- (/ (cdr collect) (cdr promote)) 1.0))
                                   #f)))))
                        '(collect promote))))
           benchmark-names)))

  (float-analysis))

; Crank all sets of derived data (across ROF settings) into one large table.

(define (rof-float-all-derived)

  (define (display-analysis data out)
    (latex-table 
     out 
     "|l|rr|rr|rr|rr|rr|"
     '(("" ("(none)" 2) ("Phase" 2) ("Phase" 2) ("Phase" 2) ("Phase" 2))
       ("" "" ""        "" ""       ("Luck"  2) ("Limit" 2) ("Luck" 2))
       ("" "" ""        "" ""       "" ""       "" ""       ("Limit" 2))
       ("" "Collection" "Promotion" "Collection" "Promotion" "Collection" 
           "Promotion" "Collection" "Promotion" "Collection" "Promotion"))
     (cons '(benchmark-name "~20a") 
           (make-list 10 '(literal "~5,1f")))
     data))

  (let ((results
         (apply map
                (lambda xs
                  (cons (car (car xs))
                        (mappend (lambda (x)
                                   (list (cadr x) (caddr x)))
                                 xs)))
                (map rof-float-analysis-derived
                     '(()
                       (-phase-detection)
                       (-phase-detection -feeling-lucky)
                       (-phase-detection -np-remset-limit)
                       (-phase-detection -feeling-lucky -np-remset-limit))))))
    
    (display-analysis results (current-output-port))))

(define (rof-float-all-derived2)

  (define (display-analysis data out)
    (latex-table 
     out 
     "|l|rrr|"
     '(("" "" ""))
     (cons '(benchmark-name "~20a") 
           (make-list 2 '(literal "~5,1f")))
     data))

  (let ((results
         (apply map
                (lambda xs
                  (cons (car (car xs))
                        (mappend (lambda (x)
                                   (list (cadr x) (caddr x)))
                                 xs)))
                (map rof-float-analysis-derived
                     '(;()
                       ;(-phase-detection)
                       ;(-phase-detection -feeling-lucky)
                       ;(-phase-detection -np-remset-limit)
                       (-phase-detection -feeling-lucky -np-remset-limit))))))
    
    (display-analysis results (current-output-port))))

            
; `All-data' as from rof-float-data
;
; Returns a list of ((benchmark . tag) . mark-volume) where mark-volume
; is averaged across all runs for the benchmark and tag is one 
; of 'collect', 'promote', 'normal'.  The 'normal' tag is from
; a run without float reduction, the 'collect' tag is from a run
; with full gc before ROF gc, and the 'promote' tag is from a run
; with full gc before every gc.

(define (rof-float-analysis-core all-data)

  (define (datum-tag datum)
    (cond ((let ((x (assq '-rof-fullgc-on-collection datum)))
             (and x (cdr x)))
           'collect)
          ((let ((x (assq '-rof-fullgc-on-promotion datum)))
             (and x (cdr x)))
           'promote)
          (else 
           'normal)))

  (define results #f)

  (define (process-datum datum n)
    (if (all-iterations-succeeded? datum)
        (let ((cell ((results 'lookup) (cons (benchmark-designator datum) 
                                             (datum-tag datum)))))
          (set-car! (cdr cell) (+ (cadr cell) 
                                  (cdr (assq 'run-bytes-copied-best datum))
                                  (cdr (assq 'run-bytes-moved-best datum))))
          (set-car! (cddr cell) (+ (caddr cell) 1)))))

  (define (analyze all-data)
    (set! results (make-benchmark-associator (lambda () 
                                               (list 0 0))))
    (for-each-benchmark-result process-datum all-data)
    (map (lambda (x) 
           (if (zero? (caddr x))
               (cons (car x) #f)
               (cons (car x) (/ (cadr x) (caddr x)))))
         ((results 'results))))

  (analyze all-data))


; ROF remset traces
;
; Produce data files for remembered set size following each minor GC and
; for incidence of major GC.

(define (rof-remset-traces)
  (process-rof-remset-traces 3.0 "plots/remset-traces/rof/000-")
  (process-rof-remset-traces 3.0 "plots/remset-traces/rof/111-"
                             '-phase-detection
                             '-feeling-lucky
                             '-np-remset-limit))

(define (process-rof-remset-traces load prefix . selectors)
  (let ((non-selectors (filter (lambda (x)
                                 (not (memq x selectors)))
                               '(-phase-detection 
                                 -feeling-lucky 
                                 -np-remset-limit))))

    (define (select-settings raw-data)
      (select-most-desirable-run-and-resort
       (flatten-option-noise
        (fluid-let ((no-warnings #t))
          (basic-statistics raw-data)))
       (lambda (a b)
         (and (every? (lambda (x) (assq x a)) selectors)
              (every? (lambda (x) (not (assq x a))) non-selectors)))))
    
    (define (doit trace dout gcout)
      (process-rtrace 
       (make-generator (lambda (return)
                         (return '(start))
                         (do ((x trace (cdr x)))
                             ((null? x))
                           (case (caar x)
                             ((rof-promote-young 
                               rof-promote-old 
                               rof-promote-both)
                              (return (list 'promote (sum (cdadar x)))))
                             ((rof-collect-old)
                              (return (list 'collect (sum (cdadar x)))))
                             (else
                              (error "Unrecognized trace record: " 
                                     (car x)))))
                         (return '(stop))
                         (return '(end))
                         (error "Read past end of trace.")))
       dout
       gcout
       #f))

    (load-rof-rtrace-results)
    (for-each (lambda (x)
                (display (benchmark-designator x))
                (display (cdr (assq 'system-info x)))
                (newline)
                (call-with-output-file 
                    (string-append
                     prefix
                     (sanitize-string
                      (format #f "~a~a-rtrace.data" 
                              (car (benchmark-designator x))
                              (cadr (benchmark-designator x)))))
                  (lambda (dout)
                    (call-with-output-file 
                        (string-append
                         prefix
                         (sanitize-string
                          (format #f "~a~a-rtrace-gc.data"
                                  (car (benchmark-designator x))
                                  (cadr (benchmark-designator x)))))
                      (lambda (gcout)
                        (doit (cdr (assq 'run-remset-trace-best x)) 
                              dout 
                              gcout))))))
              (select-load-factor 
               load 
               (select-settings (append rtrace-rof-var-1
                                        rtrace-rof-var-5))))))

; Phase detection analysis.

(define (rof-phase-data load cutoff)
  (let ((data (append (flatten-option-noise
                       (basic-statistics phase-rof-var-1))
                      (flatten-option-noise
                       (basic-statistics phase-rof-var-5)))))
    (select-load-factor
     load
     (filter (lambda (x)
               (member `(-phase-detection . ,cutoff) x))
             data))))

(define (rof-phase-mark/cons-analysis load0 . rest)
  (let ((load load0))                   ; Masks a Twobit bug (v0.48)

    (define results #f)

    (define (process-datum datum)
      (if (all-iterations-succeeded? datum)
          (let* ((bm (benchmark-designator datum))
                 (cell ((results 'lookup) bm)))
            (set-car! (cdr cell) 
                      (+ (cadr cell)
                         (/ (+ (cdr (assq 'run-bytes-copied-best datum))
                               (cdr (assq 'run-bytes-moved-best datum)))
                            (canonical-allocation bm))))
            (set-car! (cddr cell) (+ (caddr cell) 1)))))

    (define (analyze tag all-data)
      (set! results (make-benchmark-associator (lambda () (list 0 0))))
      (for-each process-datum all-data)
      (map (lambda (x) 
             (cons (cons tag (car x)) (/ (cadr x) (caddr x))))
           ((results 'results))))

    (define (mark/cons-analysis)
      (let ((results (append (analyze 0 (rof-phase-data load 0.005))
                             (analyze 1 (rof-phase-data load 0.01))
                             (analyze 2 (rof-phase-data load 0.05))
                             (analyze 3 (rof-phase-data load 0.1))
                             (analyze 4 (rof-phase-data load 0.5)))))
        (display results) (newline)
        (map (lambda (bm)
               (cons bm (map 
                         (lambda (gc)
                           (let ((v (assoc (cons gc bm) results)))
                             (if v
                                 (list gc (cdr v))
                                 (list gc #f))))
                         (iota 5))))
             benchmark-names)))

    (define (display-analysis data out)
      (latex-table
       out
       "|l|rrrrr|"
       '(("" "0.005" "0.01" "0.05" "0.1" "0.5"))
       (cons '(benchmark-name "~20a") (make-list 5 '(num-cadr "~7,5f")))
       data))

    (if (null? rest)
        (display-analysis (mark/cons-analysis) (current-output-port))
        (call-with-output-medium (car rest) 
          (lambda (out)
            (display-analysis (mark/cons-analysis) out))))))


;;;

(define (create-rof-load-vs-mark/cons-plot prefix)
  (load-rof-results)
  (create-dof-benchmark-2d-plot prefix                     ; [sic]
                                dof-effective-inverse-load ; [sic]
                                dof-mark-cons-ratio        ; [sic]
                                (let ((x (rof-canonical-data)))
                                  (list (list-ref x 0)
                                        (list-ref x 1)
                                        (list-ref x 4)
                                        (list-ref x 5)))))

(define (rof-mark-cons-plots)
  (create-rof-load-vs-mark/cons-plot "plots/rof-summary-plots/L-"))


; eof
