; Stop/copy and gen joint reports   -*- mode: scheme; mode: font-lock -*-
; 2000-07-11 / lth

(define (sc+gen-copy-cost-report)
  (let ((2gen-results (2gen-copy-cost))
        (3gen-results (3gen-copy-cost))
        (sc-results   (sc-copy-cost)))
    (latex-table
     (current-output-port)
     "|l|rr|"
     '(("" "SC" "2GEN" "3GEN"))
     `((benchmark-name "~20a") ,@(make-list 3 '(literal "~4,1f")))
     (map (lambda (bm)
            (list bm 
                  (cdr (assoc bm sc-results))
                  (cdr (assoc bm 2gen-results))
                  (cdr (assoc bm 3gen-results))))
          benchmark-names))))

(define (sc+gen-mark/cons-analysis . rest)

  (define (process-datum datum n results)
    (if (all-iterations-succeeded? datum)
        (let* ((bm (list n (cdr (assq 'benchmark-name datum))))
               (cell ((results 'lookup) bm)))
          (set-car! (cdr cell) 
                    (+ (cadr cell)
                       (/ (+ (cdr (assq 'run-bytes-copied-best datum))
                             (cdr (assq 'run-bytes-moved-best datum))
                             (cdr (assq 'run-bytes-marked-full-best datum)))
                          (canonical-allocation bm))))
          (set-car! (cddr cell) (+ (caddr cell) 1))
          (set-car! (cdddr cell) (+ (cadddr cell) 
                                    (cdr (assq 'run-major-collections-best
                                               datum)))))))

  (define (analyze tag all-data)
    (let ((results
           (make-benchmark-associator (lambda () (list 0 0 0)))))
      (for-each-benchmark-result 
       (lambda (datum n) (process-datum datum n results))
       all-data)
      (map (lambda (x) 
             (list (cons tag (car x))
                   (/ (cadr x) (caddr x))
                   (/ (cadddr x) (caddr x))))
           ((results 'results)))))

  (define (mark/cons-analysis)
    (let ((results (append (analyze 0 (sc-canonical-data))
                           (analyze 1 (2gen-canonical-data))
                           (analyze 2 (3gen-canonical-data)))))
      (map (lambda (bm)
             `(,bm 
               ,@(map 
                  (lambda (gc)
                    (let ((v (assoc (cons gc bm) results)))
                      (if v
                          (cadr v)
                          #f)))
                  (iota 3))
               ,@(map 
                  (lambda (gc)
                    (let ((v (assoc (cons gc bm) results)))
                      (if v
                          (caddr v)
                          #f)))
                  (iota 3))))
           benchmark-names)))

  (define (display-analysis data out)
    (latex-table
     out
     "|l|rrr|"
     '(("" ("Mark/cons ratio" 3) ("Major GCs" 3))
       ("" "SC" "2GEN" "3GEN" "SC" "2GEN" "3GEN"))
     `((benchmark-name "~20a") 
       ,@(make-list 3 '(literal "~4,2f"))
       ,@(make-list 3 '(literal "~5,1f")))
     data))

  (if (null? rest)
      (display-analysis (mark/cons-analysis) (current-output-port))
      (call-with-output-medium (car rest) 
        (lambda (out)
          (display-analysis (mark/cons-analysis) out)))))

(define (earley-stack-underflows canonical-data)

  (define results '())

  (define (accumulate datum n)
    (if (and (all-iterations-succeeded? datum)
             (equal? (benchmark-designator datum) '(5 earley:12)))
        (set! results (cons (list (assoc 'run-frames-flushed-best datum)
                                  (assoc 'run-frames-restored-best datum)
                                  (list-head (assoc 'system-info datum) 4))
                            results))))
        
  (for-each-benchmark-result accumulate canonical-data)
  results)

(define (stopcopy-earley-stack-underflows)
  (display-earley (earley-stack-underflows (sc-canonical-data))))

(define (2gen-earley-stack-underflows)
  (display-earley (earley-stack-underflows (2gen-canonical-data))))

(define (display-earley data)
  (for-each (lambda (x)
              (format #t "~11a ~6@a ~3@a ~3@a~%"
                      (caddr (caddr x))
                      (if (eq? 'variable (caddr (caddr x)))
                          (cadddr (caddr x))
                          (/ (round (* 10 (/ (cadddr (caddr x)) (* 1024 1024)))) 10))
                      (inexact->exact (round (/ (cdr (car x)) 1000)))
                      (inexact->exact (round (/ (cdr (cadr x)) 1000)))))
            data))

; eof
