; Reports and sundry text formatters.  -*- mode: scheme; mode: font-lock -*-
; 2000-07-24 / lth

; Recognize a mess when you see one.

(require 'list)

(define field-width 8)                  ; Override with FLUID-LET
                                        ; or other mechanisms.

(define (->string x)
  (let ((o (open-output-string)))
    (display x o)
    (get-output-string o)))

(define (tf x)                          ; "text field"
  (field-left x field-width))

(define (rtf x)                         ; "right text field"
  (field-right x field-width))

(define (nf.1 x)                        ; 1 decimal
  (if (exact? x)
      (nf x)
      (tf (number->string (/ (round (* x 10.0)) 10.0)))))

(define (rnf.1 x)                        ; 1 decimal
  (if (exact? x)
      (rnf x)
      (rtf (number->string (/ (round (* x 10.0)) 10.0)))))

(define (kf.1 x)                        ; 1 decimal
  (let loop ((x x) (e 0))
    (cond ((< x 1000)
           (if (= e 0)
               (nf.1 x)
               (tf (string-append (number->string (/ (round (* x 10.0)) 10.0))
                                  "e"
                                  (number->string e)))))
          (else
           (loop (/ x 1000) (+ e 3))))))

(define (exact.kf.1 x)
  (kf.1 (inexact->exact (round x))))

(define (rkf.1 x)                        ; 1 decimal
  (let loop ((x x) (e 0))
    (cond ((< x 1000)
           (if (= e 0)
               (rnf.1 x)
               (rtf (string-append (number->string (/ (round (* x 10.0)) 10.0))
                                   "e"
                                   (number->string e)))))
          (else
           (loop (/ x 1000) (+ e 3))))))

(define (nf.2 x)                        ; 2 decimals
  (if (exact? x)
      (nf x)
      (tf (number->string (/ (round (* x 100.0)) 100.0)))))

(define (nf.4 x)                        ; 4 decimals
  (if (exact? x)
      (nf x)
      (tf (number->string (/ (round (* x 10000.0)) 10000.0)))))

(define (rnf.2 x)                        ; 2 decimals
  (if (exact? x)
      (rnf x)
      (rtf (number->string (/ (round (* x 100.0)) 100.0)))))

(define (rnf.4 x)                        ; 4 decimals
  (if (exact? x)
      (rnf x)
      (rtf (number->string (/ (round (* x 10000.0)) 10000.0)))))

(define (pf x)                          ; Percent field
  (tf (string-append (number->string (/ (round (* x 100.0)) 100.0)) "%")))

(define (rpf x)                          ; Percent field
  (rtf (string-append (number->string (/ (round (* x 100.0)) 100.0)) "%")))

(define (nf x)                          ; "numeric field"
  (tf (number->string x)))

(define (rnf x)                          ; "numeric field"
  (rtf (number->string x)))

(define (field-left x n)                ; left-formatted field
  (if (symbol? x)
      (field-left (symbol->string x) n)
      (string-append x (make-string (- n (string-length x))))))

(define (field-right x n)               ; right-formatted field
  (if (symbol? x)
      (field-left (symbol->string x) n)
      (string-append (make-string (- n (string-length x))) x)))

(define (ms ms)                         ; format milliseconds as seconds
  (let ((s (/ (round (/ ms 10.0)) 100.0)))
    (if (and (= s 0.0) (not (= ms 0.0)))
        (tf (string-append (number->string (/ (round ms) 1000.0))
                           "s"))
        (tf (string-append (let loop ((x (number->string s)))
                             (if (char=? (string-ref x (- (string-length x) 3))
                                         #\.)
                                 x
                                 (loop (string-append x "0"))))
                           "s")))))

(define (mb b)                        ; format bytes as megabytes
  (let ((mb (/ (round (/ b (* 1024.0 512.0))) 2.0)))
    (if (< mb 1000.0)
        (tf (string-append 
             (number->string mb)
             "mb"))
        (tf (string-append 
             (number->string (inexact->exact (round mb)))
             "mb")))))

(define (kb b)                        ; format bytes as kilobytes
  (tf (string-append 
       (number->string (/ (round (/ b 512.0)) 2.0))
       "kb")))

(define (mb* b)                        ; format bytes as megabytes
  (let ((mb (/ (round (/ b (* 1024.0 512.0))) 2.0)))
    (if (< mb 1000.0)
        (tf (number->string mb))
        (tf (number->string (inexact->exact (round mb)))))))

(define (ms* ms)                         ; format milliseconds as seconds
  (let ((s (/ (round (/ ms 10.0)) 100.0)))
    (if (and (= s 0.0) (not (= ms 0.0)))
        (tf (number->string (/ (round ms) 1000.0)))
        (tf (number->string s)))))

; More rational names for the above

(define format-ms-as-s ms*)             ; no
(define format-byte-volume.1 kf.1)      ; 1 decimal and e3, e6, ...

(define find-formatter 
  (let ((formatters
         `((ms . ,ms)
           (mb . ,mb)
           (kb . ,kb)
           (bv . ,(lambda (x)           ; "boolean vector"
                    (tf (apply string-append 
                               (list-insert 
                                " " 
                                (map (lambda (x) (if x "1" "0")) x))))))
           (mb/blank . ,(lambda (n) (if (zero? n) (tf "") (mb n))))
           (bool     . ,(lambda (x) (nf (if x 1 0))))
           (bool-not . ,(lambda (x) (nf (if x 0 1))))
           (nf   . ,nf)
           (n    . ,(lambda (n) 
                      (tf (number->string 
                           (inexact->exact (truncate n))))))
           (nf.1 . ,nf.1)
           (nf.2 . ,nf.2)
           (nf.4 . ,nf.4)
           (kf.1 . ,kf.1)
           (tf   . ,tf)
           (exact.kf.1 . ,exact.kf.1))))
    (lambda (tag)
      (let ((probe (assq tag formatters)))
        (if probe
            (cdr probe)
            (error "No formatter for " tag))))))

; Print a table of statistics sorted by benchmark and system.
;
; Optional arguments
;  'filter (lambda (x) ...)    Filter the data, x is the output 
;                              from basic-statistics
;  'sort (lambda (x) ...)      Apply a final sort to the data, x is
;                              the already-sorted output from the filter
;                              function.
;  'header-filter (lambda (bm collector heap aux) ...) => (a b c d)
;                              Return four values for the header
;                              field.  Values returned as #f will not
;                              be printed.
;  'latex                      Produce a LaTeX table
;  'html                       Produce an HTML table
;  'break n                    Produce a page break after n sets
;  'break-dof-unreduced        Special minor report breaks
;  'no-warnings                Turn off warnings
;  'no-fail                    Don't print F mark, and if a
;                               run fails some iterations, print nothing
;  'no-variance                Don't print V mark
;  'no-heap-limit              Don't print heap limit field
;  'no-load/trigger            Don't print load/trigger field

(define (standard-report list-of-systems fields . rest)

  (define break-limit 100000)
  (define set-counter 0)
  (define filter-function (lambda (x) x))
  (define sorter-function (lambda (x) x))
  (define header-filter-function 
    (lambda (bm collector heap aux)
      (values bm collector heap (if (null? aux) "" aux))))
  (define break-dof-unreduced #f)
  (define warnings #t)
  (define variance #t)
  (define fail #t)
  (define heap-limit #t)
  (define load-only #t)                 ; Not yet controllable by keyword
  (define load/trigger #t)
  (define num-fixed-fields #f)

  (define fld-header car)
  (define fld-formatter cadr)
  (define fld-tag caddr)

  (define markup #f)                    ; Current markup style.
                                        ; Set during initialization

  (define (table-start) ((cdr (assq 'table-start markup))))
  (define (table-end) ((cdr (assq 'table-end markup))))
  (define (section-header bm collector heap aux)
    ((cdr (assq 'section-header markup)) bm collector heap aux))
  (define (center x last?) ((cdr (assq 'center markup)) x last?))
  (define (field-separator) (cdr (assq 'field-separator markup)))
  (define (line-begin) (cdr (assq 'line-begin markup)))
  (define (line-end) (cdr (assq 'line-end markup)))
  (define (inter-page) (cdr (assq 'inter-page markup)))

  (define text-markup
    `((table-start . ,(let ((done #f))
                        (lambda ()
                          (if (not done)
                              (begin (print-headers)
                                     (set! done #t))))))
      (table-end . ,(lambda () #t))
      (section-header 
       . ,(lambda (bm collector heap aux)
            (call-with-values
             (lambda ()
               (header-filter-function bm collector heap aux))
             (lambda xs
               (newline)
               (for-each display (list-insert " " (filter identity xs)))
               (newline)))))
      (center . ,(lambda (x last?) x))
      (field-separator . "")
      (line-begin . "  ")
      (line-end . "\n")
      (inter-page . "")))

  (define html-markup
    `((table-start . ,(lambda ()
                        (display "<p><pre>")
                        (newline)
                        (print-headers)))
      (table-end   . ,(lambda ()
                        (display "</pre></p>")
                        (newline)))
      (section-header 
       . ,(lambda (bm collector heap aux)
            (call-with-values
             (lambda ()
               (header-filter-function bm collector heap aux))
             (lambda xs
               (newline)
               (for-each display (list-insert " " (filter identity xs)))
               (newline)))))
      (center . ,(lambda (x last?) x))
      (field-separator . "")
      (line-begin . "  ")
      (line-end . "\n")
      (inter-page . "<br><br>")))

  (define latex-markup
    `((table-start . ,(lambda ()
                        (display "\\begin{tabular}{|")
                        (display (make-string (number-of-fields) #\r))
                        (display "|}")
                        (newline)
                        (display "\\hline")
                        (newline)
                        (print-headers)))
      (table-end  . ,(lambda ()
                       (display "\\hline")
                       (newline)
                       (display "\\end{tabular}")
                       (newline)))
      (section-header 
       . ,(lambda (bm collector heap aux)
            (call-with-values
             (lambda ()
               (header-filter-function bm collector heap aux))
             (lambda xs
               (format #t "\\hline~%")
               (format #t "\\multicolumn{~a}{|l|}{\\bf " (number-of-fields))
               (for-each display (list-insert " " 
                                              (map sanitize-for-latex
                                               (filter identity xs))))
               (format #t "} \\\\~%")))))
      (center . ,(lambda (x last?) 
                   (format #f "\\multicolumn{1}{c~a}{~a}" 
                           (if last? "|" "") 
                           x)))
      (field-separator . " & ")
      (line-begin . " & ")
      (line-end . " \\\\\n")
      (inter-page . "\n") ; Forces paragraph break
      ))

  (define (sanitize-for-latex x)
    (if (string? x)
        (sanitize-string-for-latex x)
        x))

  (define (number-of-fields)
    (+ (length fields)
       (if load/trigger 1 0)
       (if variance 1 0)
       (if fail 1 0)
       (if heap-limit 1 0)
       1))                              ; leading empty field

  (define (print-headers)
    (let* ((hdr   (append (map fld-header fields) 
                          (if heap-limit '(("Heap" "limit")) '())
                          (if load/trigger 
                              (if load-only 
                                  '(("Load"))
                                  '(("Load/" "Trigger")))
                              '())))
           (depth (apply max (map length hdr)))
           (width (length hdr)))
      (do ((i 0 (+ i 1))
           (hdr hdr (map (lambda (x) (if (pair? x) (cdr x) x)) hdr)))
          ((= i depth))
        (display (line-begin))
        (do ((xs (map (lambda (x) (if (pair? x) (car x) x)) hdr)
                 (cdr xs)))
            ((null? xs))
          (let ((last? (and (null? (cdr xs))
                            (not variance)
                            (not fail))))
            (let ((x (car xs)))
              (if (string? x)
                  (display (center (tf x) last?))
                  (display (center (tf "") last?)))
              (if (not (null? (cdr xs)))
                  (display (field-separator))))))
        (if variance
            (begin
              (display (field-separator))
              (if (= i 0) (display " V"))))
        (if fail
            (begin
              (display (field-separator))
              (if (= i 0) (display " F"))))
        (display (line-end)))))

  (define heading #f)                            ; previous heading printed
  (define fixed-runs '(fixed upper upper+lower)) ; tags for fixed-heap runs
  (define variable-runs '(variable))             ; tags for variable-heap runs

  (define (report-break stats)
    (let* ((sys       (cdr (assq 'system-info stats)))
           (bm        (cdr (assq 'benchmark-name stats)))
           (collector (car sys))
           (heap      (cadr sys))
           (fixed?    (memq heap fixed-runs))
           (var?      (memq heap variable-runs))
           (size      (if fixed? (caddr sys) #f))
           (trigger   (if fixed? (cadddr sys) #f))
           (load      (if var? (caddr sys) #f))
           (aux       (if fixed? (cdddr sys) (cdddr sys)))
           (x         (if break-dof-unreduced
                          (list (list size (or trigger load)) bm collector heap aux)
                          (list trigger bm collector heap aux))))
      (cond ((equal? x heading)
             'none)
            ((and heading (equal? (cdr heading) (cdr x)))
             (set! heading x)
             'minor)
            (else
             (set! heading x)
             'major))))

  ; Call only if a break is going to happen.

  (define (page-break-if-at-limit)
    (if (= set-counter break-limit)
        (begin
          (set! set-counter 0)
          (table-end)
          (display (inter-page))
          (table-start)))
    (set! set-counter (+ set-counter 1)))

  (define (new-heading stats)
    (let* ((sys       (cdr (assq 'system-info stats)))
           (bm        (benchmark-designator stats))
           (collector (car sys))
           (heap      (cadr sys))
           (fixed?    (memq heap fixed-runs))
           (var?      (memq heap variable-runs))
           (size      (if fixed? (caddr sys) #f))
           (trigger   (if fixed? (cadddr sys) #f))
           (load      (if var? (caddr sys) #f))
           (aux       (if fixed? (cdddr sys) (cdddr sys))))
      (case (report-break stats)
        ((none) #t)
        ((minor)
         (page-break-if-at-limit)
         (newline))
        ((major)
         (page-break-if-at-limit)
         (section-header bm collector heap aux))
        (else ???))
      (let* ((before (lambda () #t))
             (before (if heap-limit
                         (lambda ()
                           (before)
                           (display (field-separator))
                           (display (if size (mb size) (tf ""))))
                         before))
             (before (if load/trigger
                         (lambda ()
                           (before)
                           (display (field-separator))
                           (display (nf (if size trigger load))))
                         before)))
        before)))

  (define (print-it stats)

    (define extra (new-heading stats))  ; A thunk

    (define (get tag)
      (let ((probe (assq tag stats)))
        (if probe
            (cdr probe)
            (error "No such tag: " tag))))

    (define (present? tag)
      (if (assq tag stats) #t #f))

    (define (print-failed-run)          ; No iterations completed
      (display (apply string-append
                      (list-insert (field-separator)
                                   (make-list num-fixed-fields
                                              (make-string field-width
                                                           #\space)))))
      (extra)
      (if variance 
          (display (field-separator)))
      (if fail 
          (begin
            (display (field-separator))
            (display " 0"))))

    (define (print-successful-run)      ; At least one iteration completed
      (do ((fields fields (cdr fields)))
          ((null? fields))
        (let ((fld (car fields)))
          (let ((tag (fld-tag fld)))
            (cond ((symbol? tag)
                   (display ((find-formatter (fld-formatter fld))
                             (get tag))))
                  ((procedure? tag)
                   (display ((find-formatter (fld-formatter fld))
                             (tag get present?))))
                  (else
                   (error "Illegal field tag " tag)))
            (if (not (null? (cdr fields)))
                (display (field-separator))))))
      (extra)
      (if variance
          (begin
            (display (field-separator))
            (display (if (> (get 'elapsed-all-variance) 
                            (* 0.1 (get 'elapsed-all-average)))
                         "V"
                         " "))))
      (if fail
          (begin
            (display (field-separator))
            (if (get 'some-failed)
                (let ((probe (assq 'number-of-iterations stats)))
                  (if probe
                      (display (field-right (->string (cdr probe)) 2))
                      (display " F")))))))

    (display (line-begin))
    (if fail
        (if (not (assq 'bytes-max-heap stats))
            (print-failed-run)
            (print-successful-run))
        (if (get 'some-failed)
            (print-failed-run)
            (print-successful-run)))
    (display (line-end)))

  (set! markup text-markup)      ; Default
  (let loop ((rest rest))
    (if (not (null? rest))
        (case (car rest)
          ((filter)              (set! filter-function (cadr rest))
                                 (loop (cddr rest)))
          ((sort)                (set! sorter-function (cadr rest))
                                 (loop (cddr rest)))
          ((header-filter)       (set! header-filter-function (cadr rest))
                                 (loop (cddr rest)))
          ((no-warnings)         (set! warnings #f)
                                 (loop (cdr rest)))
          ((no-variance)         (set! variance #f)
                                 (loop (cdr rest)))
          ((no-fail)             (set! fail #f)
                                 (loop (cdr rest)))
          ((no-heap-limit)       (set! heap-limit #f)
                                 (loop (cdr rest)))
          ((no-load/trigger)     (set! load/trigger #f)
                                 (loop (cdr rest)))
          ((text)                (set! markup text-markup)
                                 (loop (cdr rest)))
          ((html)                (set! markup html-markup)
                                 (loop (cdr rest)))
          ((latex)               (set! markup latex-markup)
                                 (loop (cdr rest)))
          ((break)               (set! break-limit (cadr rest))
                                 (loop (cddr rest)))
          ((break-dof-unreduced) (set! break-dof-unreduced #t)
                                 (loop (cdr rest)))
          (else                  ???))))
  (fluid-let ((no-warnings (not warnings)))
    (set! num-fixed-fields (length (map fld-header fields)))
    (table-start)
    (for-each print-it (sorter-function
                        (sort-by-standard-header
                         (filter-function 
                          (basic-statistics list-of-systems)))))
    (table-end)))

; eof
