; $Id: prefix.sch,v 1.6 2000/02/15 18:17:47 lth Exp $
;
; This file provides an enhanced version of run-benchmark, compatible with
; both the old and new versions of run-benchmark.  Instead of printing
; the results in human-readable form, this version prints them in machine-
; readable form.
;
; Also included is a procedure 'run' that is invoked by the shell scripts
; to run a particular benchmark program.

(define (run benchmark-name iterations . args)

  (define benchmarks
    `((dummy        ,(lambda () (apply dummy-benchmark args)))
      (dynamic      ,(lambda () (apply dynamic-benchmark args)))
      (earley       ,(lambda () (apply earley-benchmark args)))
      (gcbench      ,(lambda () (apply gc-benchmark args)))
      (gcold        ,(lambda () (apply gcold args)))
      (graphs       ,(lambda () (apply graphs-benchmark args)))
      (lattice      ,(lambda () (apply lattice-benchmark args)))
      (perm         ,(lambda () (apply MpermNKL-benchmark args)))
      (nboyer       ,(lambda () (apply nboyer-benchmark args)))
      (nucleic      ,(lambda () (apply nucleic-benchmark args)))
      (pnboyer      ,(lambda () (apply pnboyer-benchmark args)))
      (psboyer      ,(lambda () (apply psboyer-benchmark args)))
      (sboyer       ,(lambda () (apply sboyer-benchmark args)))
      (softscheme   ,(lambda () (apply softscheme-benchmark args)))
      (textsearch   ,(lambda () (apply textsearch-benchmark args)))
      (twobitlong   ,(lambda () (apply twobit-benchmark 'long args)))
      (twobitshort  ,(lambda () (apply twobit-benchmark 'short args)))
      (vnboyer      ,(lambda () (apply vnboyer-benchmark args)))
      ))

  (let ((probe (assq benchmark-name benchmarks)))
    (if (not probe)
        (begin (display "ERROR The benchmark ")
               (display benchmark-name)
               (display " is not known.  Valid choices are ")
               (display (map car benchmarks))
               (exit))
        (let ((bm (cadr probe)))
          (collect)
          (do ((i 0 (+ i 1)))
              ((= i iterations))
            (bm))))))



; Adapted from Lib/Common/memstats.sch to display more numbers and to 
; display them on a machine-readable form.

; Run-benchmark
;
; (run-benchmark name thunk)
; (run-benchmark name thunk iterations)
; (run-benchmark name iterations thunk ok?)
;
; The 2/3 arg version is compatible with old Larceny releases, and is
; deprecated: a warning is printed if it is used.
;
; The 4 arg version is compatible with benchmarking code that comes 
; with Gambit-C (and superior, because it tests the final value).

(define *nesting* 0)

(define (run-benchmark . args)
  (dynamic-wind
   (lambda () (set! *nesting* (+ *nesting* 1)))
   (lambda ()
     (case (length args)
       ((2 3) (newline)
              ;(display "WARNING: Using old run-benchmark.")
              ;(newline)
              (apply old-run-benchmark args))
       ((4)   (apply new-run-benchmark args))
       (else  (error "Wrong number of arguments to run-benchmark."))))
   (lambda () (set! *nesting* (- *nesting* 1)))))

(define (new-run-benchmark name iterations thunk ok?)

  (define (loop n last-result)
    (if (zero? n)
	last-result
	(loop (- n 1) (thunk))))

  (newline)
  (display "--------------------------------------------------------")
  (newline)
  (display name)
  (newline)
  (let ((result #f))
    (run-with-stats (lambda ()
		      (set! result (loop iterations (unspecified)))))
    (if (not (ok? result))
	(error "Benchmark program returned wrong result: " result))
    (unspecified)))

(define (old-run-benchmark name thunk . rest)
  (let ((n (if (null? rest) 1 (car rest))))
    
    (define (loop n)
      (if (zero? n)
	  #t
	  (begin (thunk)
		 (loop (- n 1)))))

    (newline)
    (display "--------------------------------------------------------")
    (newline)
    (display name)
    (newline)
    (run-with-stats (lambda () (loop n)))))

(define (run-with-stats thunk)
  (if (= *nesting* 1)
      (begin
        (display "RAW-MEMSTATS-BEFORE ")
        (stats-dump-stdout)
        (newline)))
  (let ((r (thunk)))
    (if (= *nesting* 1)
        (begin
          (display "RAW-MEMSTATS-AFTER ")
          (stats-dump-stdout)
          (newline)))
    r))

(define error 
  (let ((error error))
    (lambda args
      (newline)                         ; For the benefit of parsing output
      (display "ERROR ")
      (apply error args))))

; eof

